import s3fs


def create_s3_file_system(s3_client_kwargs: dict) -> s3fs.S3FileSystem:
    if not s3_client_kwargs:
        return s3fs.S3FileSystem(anon=True)

    config_kwargs = {}
    if s3_client_kwargs.get("config") is not None:
        boto_config = s3_client_kwargs.pop("config")
        for key, val in boto_config.__dict__.items():
            if not key.startswith("_") and val is not None:
                config_kwargs[key] = val

    anon = False
    if s3_client_kwargs.get("aws_access_key_id") is None:
        anon = True

    return s3fs.S3FileSystem(
        anon=anon, client_kwargs=s3_client_kwargs, config_kwargs=config_kwargs or None
    )



from typing import Any, Callable, Dict, List, Optional, Set, Union

import pyarrow as pa

from deltacat.storage import (
    Delta,
    DeltaLocator,
    DeltaType,
    DistributedDataset,
    LifecycleState,
    ListResult,
    LocalDataset,
    LocalTable,
    Manifest,
    ManifestAuthor,
    Namespace,
    Partition,
    SchemaConsistencyType,
    Stream,
    StreamLocator,
    Table,
    TableVersion,
    SortKey,
    PartitionLocator,
)
from deltacat.types.media import ContentType, StorageType, TableType
from deltacat.utils.common import ReadKwargsProvider


def list_namespaces(*args, **kwargs) -> ListResult[Namespace]:
    """
    Lists a page of table namespaces. Namespaces are returned as list result
    items.
    """
    raise NotImplementedError("list_namespaces not implemented")


def list_tables(namespace: str, *args, **kwargs) -> ListResult[Table]:
    """
    Lists a page of tables for the given table namespace. Tables are returned as
    list result items. Raises an error if the given namespace does not exist.
    """
    raise NotImplementedError("list_tables not implemented")


def list_table_versions(
    namespace: str, table_name: str, *args, **kwargs
) -> ListResult[TableVersion]:
    """
    Lists a page of table versions for the given table. Table versions are
    returned as list result items. Raises an error if the given table does not
    exist.
    """
    raise NotImplementedError("list_table_versions not implemented")


def list_partitions(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs
) -> ListResult[Partition]:
    """
    Lists a page of partitions for the given table version. Partitions are
    returned as list result items. Table version resolves to the latest active
    table version if not specified. Raises an error if the table version does
    not exist.
    """
    raise NotImplementedError("list_partitions not implemented")


def list_stream_partitions(stream: Stream, *args, **kwargs) -> ListResult[Partition]:
    """
    Lists all partitions committed to the given stream.
    """
    raise NotImplementedError("list_stream_partitions not implemented")


def list_deltas(
    namespace: str,
    table_name: str,
    partition_values: Optional[List[Any]] = None,
    table_version: Optional[str] = None,
    first_stream_position: Optional[int] = None,
    last_stream_position: Optional[int] = None,
    ascending_order: Optional[bool] = None,
    include_manifest: bool = False,
    *args,
    **kwargs
) -> ListResult[Delta]:
    """
    Lists a page of deltas for the given table version and committed partition.
    Deltas are returned as list result items. Deltas returned can optionally be
    limited to inclusive first and last stream positions. Deltas are returned by
    descending stream position by default. Table version resolves to the latest
    active table version if not specified. Partition values should not be
    specified for unpartitioned tables. Raises an error if the given table
    version or partition does not exist.

    To conserve memory, the deltas returned do not include manifests by
    default. The manifests can either be optionally retrieved as part of this
    call or lazily loaded via subsequent calls to `get_delta_manifest`.
    """
    raise NotImplementedError("list_deltas not implemented")


def list_partition_deltas(
    partition_like: Union[Partition, PartitionLocator],
    first_stream_position: Optional[int] = None,
    last_stream_position: Optional[int] = None,
    ascending_order: bool = False,
    include_manifest: bool = False,
    *args,
    **kwargs
) -> ListResult[Delta]:
    """
    Lists a page of deltas committed to the given partition.

    To conserve memory, the deltas returned do not include manifests by
    default. The manifests can either be optionally retrieved as part of this
    call or lazily loaded via subsequent calls to `get_delta_manifest`.
    """
    raise NotImplementedError("list_partition_deltas not implemented")


def get_delta(
    namespace: str,
    table_name: str,
    stream_position: int,
    partition_values: Optional[List[Any]] = None,
    table_version: Optional[str] = None,
    include_manifest: bool = False,
    *args,
    **kwargs
) -> Optional[Delta]:
    """
    Gets the delta for the given table version, partition, and stream position.
    Table version resolves to the latest active table version if not specified.
    Partition values should not be specified for unpartitioned tables. Raises
    an error if the given table version or partition does not exist.

    To conserve memory, the delta returned does not include a manifest by
    default. The manifest can either be optionally retrieved as part of this
    call or lazily loaded via a subsequent call to `get_delta_manifest`.
    """
    raise NotImplementedError("get_delta not implemented")


def get_latest_delta(
    namespace: str,
    table_name: str,
    partition_values: Optional[List[Any]] = None,
    table_version: Optional[str] = None,
    include_manifest: bool = False,
    *args,
    **kwargs
) -> Optional[Delta]:
    """
    Gets the latest delta (i.e. the delta with the greatest stream position) for
    the given table version and partition. Table version resolves to the latest
    active table version if not specified. Partition values should not be
    specified for unpartitioned tables. Raises an error if the given table
    version or partition does not exist.

    To conserve memory, the delta returned does not include a manifest by
    default. The manifest can either be optionally retrieved as part of this
    call or lazily loaded via a subsequent call to `get_delta_manifest`.
    """
    raise NotImplementedError("get_latest_delta not implemented")


def download_delta(
    delta_like: Union[Delta, DeltaLocator],
    table_type: TableType = TableType.PYARROW,
    storage_type: StorageType = StorageType.DISTRIBUTED,
    max_parallelism: Optional[int] = None,
    columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
    ray_options_provider: Callable[[int, Any], Dict[str, Any]] = None,
    *args,
    **kwargs
) -> Union[LocalDataset, DistributedDataset]:
    """
    Download the given delta or delta locator into either a list of
    tables resident in the local node's memory, or into a dataset distributed
    across this Ray cluster's object store memory. Ordered table N of a local
    table list, or ordered block N of a distributed dataset, always contain
    the contents of ordered delta manifest entry N.
    """
    raise NotImplementedError("download_delta not implemented")


def download_delta_manifest_entry(
    delta_like: Union[Delta, DeltaLocator],
    entry_index: int,
    table_type: TableType = TableType.PYARROW,
    columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
    *args,
    **kwargs
) -> LocalTable:
    """
    Downloads a single manifest entry into the specified table type for the
    given delta or delta locator. If a delta is provided with a non-empty
    manifest, then the entry is downloaded from this manifest. Otherwise, the
    manifest is first retrieved then the given entry index downloaded.
    """
    raise NotImplementedError("download_delta_manifest_entry not implemented")


def get_delta_manifest(
    delta_like: Union[Delta, DeltaLocator], *args, **kwargs
) -> Manifest:
    """
    Get the manifest associated with the given delta or delta locator. This
    always retrieves the authoritative remote copy of the delta manifest, and
    never the local manifest defined for any input delta.
    """
    raise NotImplementedError("get_delta_manifest not implemented")


def create_namespace(
    namespace: str, permissions: Dict[str, Any], *args, **kwargs
) -> Namespace:
    """
    Creates a table namespace with the given name and permissions. Returns
    the created namespace.
    """
    raise NotImplementedError("create_namespace not implemented")


def update_namespace(
    namespace: str,
    permissions: Optional[Dict[str, Any]] = None,
    new_namespace: Optional[str] = None,
    *args,
    **kwargs
) -> None:
    """
    Updates a table namespace's name and/or permissions. Raises an error if the
    given namespace does not exist.
    """
    raise NotImplementedError("update_namespace not implemented")


def create_table_version(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    schema: Optional[Union[pa.Schema, str, bytes]] = None,
    schema_consistency: Optional[Dict[str, SchemaConsistencyType]] = None,
    partition_keys: Optional[List[Dict[str, Any]]] = None,
    primary_key_column_names: Optional[Set[str]] = None,
    sort_keys: Optional[List[SortKey]] = None,
    table_version_description: Optional[str] = None,
    table_version_properties: Optional[Dict[str, str]] = None,
    table_permissions: Optional[Dict[str, Any]] = None,
    table_description: Optional[str] = None,
    table_properties: Optional[Dict[str, str]] = None,
    supported_content_types: Optional[List[ContentType]] = None,
    *args,
    **kwargs
) -> Stream:
    """
    Create a table version with an unreleased lifecycle state and an empty delta
    stream. Table versions may be schemaless and unpartitioned, or partitioned
    according to a list of partition key names and types. Note that partition
    keys are not required to exist in the table's schema, and can thus still be
    used with schemaless tables. This can be useful for creating logical shards
    of a delta stream where partition keys are known but not projected onto each
    row of the table (e.g. all rows of a customer orders table are known to
    correspond to a given order day, even if this column doesn't exist in the
    table). Primary and sort keys must exist within the table's schema.
    Permissions specified at the table level override any conflicting
    permissions specified at the table namespace level. Returns the stream
    for the created table version. Raises an error if the given namespace does
    not exist.

    Schemas are optional for DeltaCAT tables and can be used to inform the data
    consistency checks run for each field. If a schema is present, it can be
    used to enforce the following column-level data consistency policies at
    table load time:

    None: No consistency checks are run. May be mixed with the below two
    policies by specifying column names to pass through together with
    column names to coerce/validate.

    Coerce: Coerce fields to fit the schema whenever possible. An explicit
    subset of column names to coerce may optionally be specified.

    Validate: Raise an error for any fields that don't fit the schema. An
    explicit subset of column names to validate may optionally be specified.
    """
    raise NotImplementedError("create_table_version not implemented")


def update_table(
    namespace: str,
    table_name: str,
    permissions: Optional[Dict[str, Any]] = None,
    description: Optional[str] = None,
    properties: Optional[Dict[str, str]] = None,
    new_table_name: Optional[str] = None,
    *args,
    **kwargs
) -> None:
    """
    Update table metadata describing the table versions it contains. By default,
    a table's properties are empty, and its description and permissions are
    equal to those given when its first table version was created. Raises an
    error if the given table does not exist.
    """
    raise NotImplementedError("update_table not implemented")


def update_table_version(
    namespace: str,
    table_name: str,
    table_version: str,
    lifecycle_state: Optional[LifecycleState] = None,
    schema: Optional[Union[pa.Schema, str, bytes]] = None,
    schema_consistency: Optional[Dict[str, SchemaConsistencyType]] = None,
    description: Optional[str] = None,
    properties: Optional[Dict[str, str]] = None,
    *args,
    **kwargs
) -> None:
    """
    Update a table version. Notably, updating an unreleased table version's
    lifecycle state to 'active' telegraphs that it is ready for external
    consumption, and causes all calls made to consume/produce streams,
    partitions, or deltas from/to its parent table to automatically resolve to
    this table version by default (i.e. when the client does not explicitly
    specify a different table version). Raises an error if the given table
    version does not exist.
    """
    raise NotImplementedError("update_table_version not implemented")


def stage_stream(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs
) -> Stream:
    """
    Stages a new delta stream for the given table version. Resolves to the
    latest active table version if no table version is given. Returns the
    staged stream. Raises an error if the table version does not exist.
    """
    raise NotImplementedError("stage_stream not implemented")


def commit_stream(stream: Stream, *args, **kwargs) -> Stream:
    """
    Registers a delta stream with a target table version, replacing any
    previous stream registered for the same table version. Returns the
    committed stream.
    """
    raise NotImplementedError("commit_stream not implemented")


def delete_stream(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs
) -> None:
    """
    Deletes the delta stream currently registered with the given table version.
    Resolves to the latest active table version if no table version is given.
    Raises an error if the table version does not exist.
    """
    raise NotImplementedError("delete_stream not implemented")


def get_stream(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs
) -> Optional[Stream]:
    """
    Gets the most recently committed stream for the given table version and
    partition key values. Resolves to the latest active table version if no
    table version is given. Returns None if the table version does not exist.
    """
    raise NotImplementedError("get_stream not implemented")


def stage_partition(
    stream: Stream, partition_values: Optional[List[Any]] = None, *args, **kwargs
) -> Partition:
    """
    Stages a new partition for the given stream and partition values. Returns
    the staged partition. If this partition will replace another partition
    with the same partition values, then it will have its previous partition ID
    set to the ID of the partition being replaced. Partition keys should not be
    specified for unpartitioned tables.
    """
    raise NotImplementedError("stage_partition not implemented")


def commit_partition(partition: Partition, *args, **kwargs) -> Partition:
    """
    Commits the given partition to its associated table version stream,
    replacing any previous partition registered for the same stream and
    partition values. Returns the registered partition. If the partition's
    previous delta stream position is specified, then the commit will
    be rejected if it does not match the actual previous stream position of
    the partition being replaced. If the partition's previous partition ID is
    specified, then the commit will be rejected if it does not match the actual
    ID of the partition being replaced.
    """
    raise NotImplementedError("commit_partition not implemented")


def delete_partition(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    partition_values: Optional[List[Any]] = None,
    *args,
    **kwargs
) -> None:
    """
    Deletes the given partition from the specified table version. Resolves to
    the latest active table version if no table version is given. Partition
    values should not be specified for unpartitioned tables. Raises an error
    if the table version or partition does not exist.
    """
    raise NotImplementedError("delete_partition not implemented")


def get_partition(
    stream_locator: StreamLocator,
    partition_values: Optional[List[Any]] = None,
    *args,
    **kwargs
) -> Optional[Partition]:
    """
    Gets the most recently committed partition for the given stream locator and
    partition key values. Returns None if no partition has been committed for
    the given table version and/or partition key values. Partition values
    should not be specified for unpartitioned tables.
    """
    raise NotImplementedError("get_partition not implemented")


def stage_delta(
    data: Union[LocalTable, LocalDataset, DistributedDataset, Manifest],
    partition: Partition,
    delta_type: DeltaType = DeltaType.UPSERT,
    max_records_per_entry: Optional[int] = None,
    author: Optional[ManifestAuthor] = None,
    properties: Optional[Dict[str, str]] = None,
    s3_table_writer_kwargs: Optional[Dict[str, Any]] = None,
    content_type: ContentType = ContentType.PARQUET,
    *args,
    **kwargs
) -> Delta:
    """
    Writes the given table to 1 or more S3 files. Returns an unregistered
    delta whose manifest entries point to the uploaded files. Applies any
    schema consistency policies configured for the parent table version.
    """
    raise NotImplementedError("stage_delta not implemented")


def commit_delta(delta: Delta, *args, **kwargs) -> Delta:
    """
    Registers a new delta with its associated target table version and
    partition. Returns the registered delta. If the delta's previous stream
    position is specified, then the commit will be rejected if it does not match
    the target partition's actual previous stream position. If the delta's
    stream position is specified, it must be greater than the latest stream
    position in the target partition.
    """
    raise NotImplementedError("commit_delta not implemented")


def get_namespace(namespace: str, *args, **kwargs) -> Optional[Namespace]:
    """
    Gets table namespace metadata for the specified table namespace. Returns
    None if the given namespace does not exist.
    """
    raise NotImplementedError("get_namespace not implemented")


def namespace_exists(namespace: str, *args, **kwargs) -> bool:
    """
    Returns True if the given table namespace exists, False if not.
    """
    raise NotImplementedError("namespace_exists not implemented")


def get_table(namespace: str, table_name: str, *args, **kwargs) -> Optional[Table]:
    """
    Gets table metadata for the specified table. Returns None if the given
    table does not exist.
    """
    raise NotImplementedError("get_table not implemented")


def table_exists(namespace: str, table_name: str, *args, **kwargs) -> bool:
    """
    Returns True if the given table exists, False if not.
    """
    raise NotImplementedError("table_exists not implemented")


def get_table_version(
    namespace: str, table_name: str, table_version: str, *args, **kwargs
) -> Optional[TableVersion]:
    """
    Gets table version metadata for the specified table version. Returns None
    if the given table version does not exist.
    """
    raise NotImplementedError("get_table_version not implemented")


def get_latest_table_version(
    namespace: str, table_name: str, *args, **kwargs
) -> Optional[TableVersion]:
    """
    Gets table version metadata for the latest version of the specified table.
    Returns None if no table version exists for the given table.
    """
    raise NotImplementedError("get_latest_table_version not implemented")


def get_latest_active_table_version(
    namespace: str, table_name: str, *args, **kwargs
) -> Optional[TableVersion]:
    """
    Gets table version metadata for the latest active version of the specified
    table. Returns None if no active table version exists for the given table.
    """
    raise NotImplementedError("get_latest_active_table_version not implemented")


def get_table_version_column_names(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs
) -> Optional[List[str]]:
    """
    Gets a list of column names for the specified table version, or for the
    latest active table version if none is specified. The index of each
    column name returned represents its ordinal position in a delimited text
    file or other row-oriented content type files appended to the table.
    Returns None for schemaless tables. Raises an error if the table version
    does not exist.
    """
    raise NotImplementedError("get_table_version_column_names not implemented")


def get_table_version_schema(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs
) -> Optional[Union[pa.Schema, str, bytes]]:
    """
    Gets the schema for the specified table version, or for the latest active
    table version if none is specified. Returns None if the table version is
    schemaless. Raises an error if the table version does not exist.
    """
    raise NotImplementedError("get_table_version_schema not implemented")


def table_version_exists(
    namespace: str, table_name: str, table_version: str, *args, **kwargs
) -> bool:
    """
    Returns True if the given table version exists, False if not.
    """
    raise NotImplementedError("table_version_exists not implemented")



from collections import Counter

import ray


@ray.remote
class DistributedCounter(object):
    """Distributed Ray Actor wrapper around a standard collections Counter."""

    def __init__(self, *args, **kwargs):
        self.counter = Counter(*args, **kwargs)

    def increment_counter(self, other: Counter):
        self.counter += other

    def decrement_counter(self, other: Counter):
        self.counter -= other

    def intersection(self, other: Counter):
        self.counter &= other

    def union(self, other: Counter):
        self.counter |= other

    def increment(self, key, delta):
        self.counter[key] += delta

    def decrement(self, key, delta):
        self.counter[key] -= delta

    def negate(self, key):
        self.counter[key] = -self.counter[key]

    def divide(self, key, divisor):
        self.counter[key] /= divisor

    def multiply(self, key, multiplier):
        self.counter[key] *= multiplier

    def pow(self, key, exponent):
        self.counter[key] **= exponent

    def most_common(self, n=None):
        return self.counter.most_common(n)

    def elements(self):
        return self.counter.elements()

    def update(self, *args, **kwargs):
        self.counter.update(*args, **kwargs)

    def subtract(self, *args, **kwargs):
        self.counter.subtract(*args, **kwargs)

    def clear(self):
        self.counter.clear()

    def counter(self):
        return self.counter

    def get(self, key):
        return self.counter[key]

    def set(self, key, value):
        self.counter[key] = value

    def setdefault(self, key, default):
        return self.counter.setdefault(key, default)

    def pop(self, key):
        return self.counter.pop(key)

    def popitem(self):
        return self.counter.popitem()

    def items(self):
        return self.counter.items()

    def keys(self):
        return self.counter.keys()

    def values(self):
        return self.counter.values()



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

from typing import Any, Callable, Dict, Optional, TypeVar, Union, cast

import pyarrow as pa
import s3fs
from ray.data import Dataset
from ray.data.datasource import BlockWritePathProvider, DefaultBlockWritePathProvider

from deltacat.io.aws.redshift.redshift_datasource import RedshiftDatasource

T = TypeVar("T")


class DeltacatDataset(Dataset[T]):
    @staticmethod
    def from_dataset(dataset: Dataset[T]) -> DeltacatDataset[T]:
        # cast to DeltacatDataset in-place since it only adds new methods
        dataset.__class__ = DeltacatDataset
        return cast(DeltacatDataset[T], dataset)

    def write_redshift(
        self,
        path: str,
        *,
        filesystem: Optional[Union[pa.fs.FileSystem, s3fs.S3FileSystem]] = None,
        try_create_dir: bool = True,
        arrow_open_stream_args: Optional[Dict[str, Any]] = None,
        block_path_provider: BlockWritePathProvider = DefaultBlockWritePathProvider(),
        arrow_parquet_args_fn: Callable[[], Dict[str, Any]] = lambda: {},
        **arrow_parquet_args,
    ) -> None:
        """Writes the dataset to Parquet files and commits a Redshift manifest
        back to S3 indexing the files written. The output can be loaded into
        Redshift by providing it to the Redshift COPY command, or via AWS Data
        Wrangler's `wr.redshift.copy_from_files()` API.

        This is only supported for datasets convertible to Arrow records.
        To control the number of files, use ``.repartition()``.

        Unless a custom block path provider is given, the format of the output
        files will be {uuid}_{block_idx}.parquet, where ``uuid`` is an unique
        id for the dataset.

        The Redshift manifest will be written to ``f"{path}/manifest``

        Examples:
            >>> ds.write_redshift("s3://bucket/path")

        Time complexity: O(dataset size / parallelism)

        Args:
            path: The path to the destination root directory where Parquet
                files and the Redshift manifest will be written to.
            filesystem: The filesystem implementation to write to. This should
                be either PyArrow's S3FileSystem or s3fs.
            try_create_dir: Try to create all directories in destination path
                if True. Does nothing if all directories already exist.
            arrow_open_stream_args: kwargs passed to
                pyarrow.fs.FileSystem.open_output_stream
            block_path_provider: BlockWritePathProvider implementation
                to write each dataset block to a custom output path. Uses
                DefaultBlockWritePathProvider if None.
            arrow_parquet_args_fn: Callable that returns a dictionary of write
                arguments to use when writing each block to a file. Overrides
                any duplicate keys from arrow_parquet_args. This should be used
                instead of arrow_parquet_args if any of your write arguments
                cannot be pickled, or if you'd like to lazily resolve the write
                arguments for each dataset block.
            arrow_parquet_args: Options to pass to
                pyarrow.parquet.write_table(), which is used to write out each
                block to a file.
        """
        self.write_datasource(
            RedshiftDatasource(),
            path=path,
            dataset_uuid=self._uuid,
            filesystem=filesystem,
            try_create_dir=try_create_dir,
            open_stream_args=arrow_open_stream_args,
            block_path_provider=block_path_provider,
            write_args_fn=arrow_parquet_args_fn,
            **arrow_parquet_args,
        )



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations
from typing import Any, Dict, List, Optional, Set, Tuple
import pyarrow as pa

from deltacat.tests.compute.test_util_common import (
    PartitionKey,
)

from deltacat.storage import (
    Delta,
    DeltaType,
    Partition,
    PartitionLocator,
    Stream,
)
from deltacat.tests.compute.test_util_common import (
    create_src_table,
    create_destination_table,
    create_rebase_table,
)


def create_incremental_deltas_on_source_table(
    source_namespace: str,
    source_table_name: str,
    source_table_version: str,
    source_table_stream: Stream,
    partition_values_param,
    incremental_deltas: pa.Table,
    incremental_delta_type: DeltaType,
    ds_mock_kwargs: Optional[Dict[str, Any]] = None,
) -> Tuple[PartitionLocator, Delta]:
    import deltacat.tests.local_deltacat_storage as ds

    src_partition: Partition = ds.get_partition(
        source_table_stream.locator,
        partition_values_param,
        **ds_mock_kwargs,
    )
    new_delta: Delta = ds.commit_delta(
        ds.stage_delta(
            incremental_deltas, src_partition, incremental_delta_type, **ds_mock_kwargs
        ),
        **ds_mock_kwargs,
    )
    src_table_stream_after_committed_delta: Stream = ds.get_stream(
        source_namespace,
        source_table_name,
        source_table_version,
        **ds_mock_kwargs,
    )
    src_partition_after_committed_delta: Partition = ds.get_partition(
        src_table_stream_after_committed_delta.locator,
        partition_values_param,
        **ds_mock_kwargs,
    )
    return src_partition_after_committed_delta.locator, new_delta


def create_src_w_deltas_destination_plus_destination(
    primary_keys: Set[str],
    sort_keys: Optional[List[Any]],
    partition_keys: Optional[List[PartitionKey]],
    input_deltas: pa.Table,
    input_delta_type: DeltaType,
    partition_values: Optional[List[Any]],
    ds_mock_kwargs: Optional[Dict[str, Any]],
) -> Tuple[Stream, Stream, Optional[Stream]]:
    import deltacat.tests.local_deltacat_storage as ds

    source_namespace, source_table_name, source_table_version = create_src_table(
        primary_keys, sort_keys, partition_keys, ds_mock_kwargs
    )

    source_table_stream: Stream = ds.get_stream(
        namespace=source_namespace,
        table_name=source_table_name,
        table_version=source_table_version,
        **ds_mock_kwargs,
    )
    staged_partition: Partition = ds.stage_partition(
        source_table_stream, partition_values, **ds_mock_kwargs
    )
    ds.commit_delta(
        ds.stage_delta(
            input_deltas, staged_partition, input_delta_type, **ds_mock_kwargs
        ),
        **ds_mock_kwargs,
    )
    ds.commit_partition(staged_partition, **ds_mock_kwargs)
    source_table_stream_after_committed: Stream = ds.get_stream(
        namespace=source_namespace,
        table_name=source_table_name,
        table_version=source_table_version,
        **ds_mock_kwargs,
    )
    (
        destination_table_namespace,
        destination_table_name,
        destination_table_version,
    ) = create_destination_table(
        primary_keys, sort_keys, partition_keys, ds_mock_kwargs
    )
    destination_table_stream: Stream = ds.get_stream(
        namespace=destination_table_namespace,
        table_name=destination_table_name,
        table_version=destination_table_version,
        **ds_mock_kwargs,
    )
    return source_table_stream_after_committed, destination_table_stream, None


def create_src_w_deltas_destination_rebase_w_deltas_strategy(
    primary_keys: Set[str],
    sort_keys: Optional[List[Any]],
    partition_keys: Optional[List[PartitionKey]],
    input_deltas: pa.Table,
    input_delta_type: DeltaType,
    partition_values: Optional[List[Any]],
    ds_mock_kwargs: Optional[Dict[str, Any]],
) -> Tuple[Stream, Stream, Optional[Stream]]:
    import deltacat.tests.local_deltacat_storage as ds
    from deltacat.storage import Partition, Stream

    source_namespace, source_table_name, source_table_version = create_src_table(
        primary_keys, sort_keys, partition_keys, ds_mock_kwargs
    )

    source_table_stream: Stream = ds.get_stream(
        namespace=source_namespace,
        table_name=source_table_name,
        table_version=source_table_version,
        **ds_mock_kwargs,
    )
    staged_partition: Partition = ds.stage_partition(
        source_table_stream, partition_values, **ds_mock_kwargs
    )
    ds.commit_delta(
        ds.stage_delta(
            input_deltas, staged_partition, input_delta_type, **ds_mock_kwargs
        ),
        **ds_mock_kwargs,
    )
    ds.commit_partition(staged_partition, **ds_mock_kwargs)
    source_table_stream_after_committed: Stream = ds.get_stream(
        namespace=source_namespace,
        table_name=source_table_name,
        table_version=source_table_version,
        **ds_mock_kwargs,
    )
    # create the destination table
    (
        destination_table_namespace,
        destination_table_name,
        destination_table_version,
    ) = create_destination_table(
        primary_keys, sort_keys, partition_keys, ds_mock_kwargs
    )
    # create the rebase table
    (
        rebase_table_namespace,
        rebase_table_name,
        rebase_table_version,
    ) = create_rebase_table(primary_keys, sort_keys, partition_keys, ds_mock_kwargs)
    rebasing_table_stream: Stream = ds.get_stream(
        namespace=rebase_table_namespace,
        table_name=rebase_table_name,
        table_version=rebase_table_version,
        **ds_mock_kwargs,
    )
    staged_partition: Partition = ds.stage_partition(
        rebasing_table_stream, partition_values, **ds_mock_kwargs
    )
    ds.commit_delta(
        ds.stage_delta(input_deltas, staged_partition, **ds_mock_kwargs),
        **ds_mock_kwargs,
    )
    ds.commit_partition(staged_partition, **ds_mock_kwargs)

    # get streams
    destination_table_stream: Stream = ds.get_stream(
        namespace=destination_table_namespace,
        table_name=destination_table_name,
        table_version=destination_table_version,
        **ds_mock_kwargs,
    )
    rebased_stream_after_committed: Stream = ds.get_stream(
        namespace=rebase_table_namespace,
        table_name=rebase_table_name,
        table_version=rebase_table_version,
        **ds_mock_kwargs,
    )
    return (
        source_table_stream_after_committed,
        destination_table_stream,
        rebased_stream_after_committed,
    )



# Default to use r5.8xlarge instance type for stats collection cluster
STATS_CLUSTER_R5_INSTANCE_TYPE = 8
# Using R5 instance type, 8GiB memory is available per cpu
R5_MEMORY_PER_CPU = 8
# Default to use r5.8xlarge instance type for stats collection cluster
DEFAULT_CPUS_PER_INSTANCE_R5_8XLARGE = 32
# memory reserved for head node object store
HEAD_NODE_OBJECT_STORE_MEMORY_RESERVE_RATIO = 0.3
# memory reserved for worker node object store
WORKER_NODE_OBJECT_STORE_MEMORY_RESERVE_RATIO = 0.1
# each cpu should not be processing more than this number of files to avoid unreasonable S3 I/O latency
MANIFEST_FILE_COUNT_PER_CPU = 200
# MAX_WORKER_MULTIPLIER * min_workers = max_workers to determine max workers based on min workers given
MAX_WORKER_MULTIPLIER = 2
# default trace id used for metastats collection triggered without trace id
DEFAULT_JOB_RUN_TRACE_ID = "0"



import importlib
import logging
import time
from uuid import uuid4
from collections import defaultdict
from contextlib import nullcontext
from itertools import chain, repeat
from typing import List, Optional, Tuple, Dict, Any
import pyarrow as pa
import numpy as np
import ray
from deltacat import logs
from deltacat.compute.compactor import (
    MaterializeResult,
    PyArrowWriteResult,
    RoundCompletionInfo,
)
from deltacat.compute.compactor.steps.dedupe import (
    DedupeTaskIndexWithObjectId,
    DeltaFileLocatorToRecords,
)
from deltacat.storage import (
    Delta,
    DeltaLocator,
    DeltaType,
    Partition,
    PartitionLocator,
    Manifest,
    ManifestEntry,
)
from deltacat.storage import interface as unimplemented_deltacat_storage
from deltacat.utils.common import ReadKwargsProvider
from deltacat.types.media import DELIMITED_TEXT_CONTENT_TYPES, ContentType
from deltacat.types.tables import TABLE_CLASS_TO_SIZE_FUNC
from deltacat.utils.performance import timed_invocation
from deltacat.io.object_store import IObjectStore
from deltacat.utils.pyarrow import (
    ReadKwargsProviderPyArrowCsvPureUtf8,
    ReadKwargsProviderPyArrowSchemaOverride,
    RecordBatchTables,
)
from deltacat.utils.ray_utils.runtime import (
    get_current_ray_task_id,
    get_current_ray_worker_id,
)
from deltacat.utils.metrics import emit_timer_metrics, MetricsConfig
from deltacat.utils.resources import get_current_node_peak_memory_usage_in_bytes

if importlib.util.find_spec("memray"):
    import memray

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


@ray.remote
def materialize(
    source_partition_locator: PartitionLocator,
    round_completion_info: Optional[RoundCompletionInfo],
    partition: Partition,
    mat_bucket_index: int,
    dedupe_task_idx_and_obj_id_tuples: List[DedupeTaskIndexWithObjectId],
    max_records_per_output_file: int,
    compacted_file_content_type: ContentType,
    enable_manifest_entry_copy_by_reference: bool,
    enable_profiler: bool,
    metrics_config: MetricsConfig,
    schema: Optional[pa.Schema] = None,
    read_kwargs_provider: Optional[ReadKwargsProvider] = None,
    s3_table_writer_kwargs: Optional[Dict[str, Any]] = None,
    object_store: Optional[IObjectStore] = None,
    deltacat_storage=unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Optional[Dict[str, Any]] = None,
):
    if deltacat_storage_kwargs is None:
        deltacat_storage_kwargs = {}

    def _stage_delta_from_manifest_entry_reference_list(
        manifest_entry_list_reference: List[ManifestEntry],
        partition: Partition,
        delta_type: DeltaType = DeltaType.UPSERT,
    ) -> Delta:
        assert (
            delta_type == DeltaType.UPSERT
        ), "Stage delta with existing manifest entries only supports UPSERT delta type!"
        manifest = Manifest.of(entries=manifest_entry_list_reference, uuid=str(uuid4()))
        delta = Delta.of(
            locator=DeltaLocator.of(partition.locator),
            delta_type=delta_type,
            meta=manifest.meta,
            manifest=manifest,
            previous_stream_position=partition.stream_position,
            properties={},
        )
        return delta

    # TODO (rkenmi): Add docstrings for the steps in the compaction workflow
    #  https://github.com/ray-project/deltacat/issues/79
    def _materialize(compacted_tables: List[pa.Table]) -> MaterializeResult:
        compacted_table = pa.concat_tables(compacted_tables)
        if compacted_file_content_type in DELIMITED_TEXT_CONTENT_TYPES:
            # TODO (rkenmi): Investigate if we still need to convert this table to pandas DataFrame
            # TODO (pdames): compare performance to pandas-native materialize path
            df = compacted_table.to_pandas(split_blocks=True, self_destruct=True)
            compacted_table = df
        delta, stage_delta_time = timed_invocation(
            deltacat_storage.stage_delta,
            compacted_table,
            partition,
            max_records_per_entry=max_records_per_output_file,
            content_type=compacted_file_content_type,
            s3_table_writer_kwargs=s3_table_writer_kwargs,
            **deltacat_storage_kwargs,
        )
        compacted_table_size = TABLE_CLASS_TO_SIZE_FUNC[type(compacted_table)](
            compacted_table
        )
        logger.debug(
            f"Time taken for materialize task"
            f" to upload {len(compacted_table)} records"
            f" of size {compacted_table_size} is: {stage_delta_time}s"
        )
        manifest = delta.manifest
        manifest_records = manifest.meta.record_count
        assert manifest_records == len(compacted_table), (
            f"Unexpected Error: Materialized delta manifest record count "
            f"({manifest_records}) does not equal compacted table record count "
            f"({len(compacted_table)})"
        )
        materialize_result = MaterializeResult.of(
            delta=delta,
            task_index=mat_bucket_index,
            # TODO (pdames): Generalize WriteResult to contain in-memory-table-type
            #  and in-memory-table-bytes instead of tight coupling to paBytes
            pyarrow_write_result=PyArrowWriteResult.of(
                len(manifest.entries),
                TABLE_CLASS_TO_SIZE_FUNC[type(compacted_table)](compacted_table),
                manifest.meta.content_length,
                len(compacted_table),
            ),
        )
        logger.info(f"Materialize result: {materialize_result}")
        return materialize_result

    logger.info(
        f"Starting materialize task with"
        f" materialize bucket index: {mat_bucket_index}..."
    )
    task_id = get_current_ray_task_id()
    worker_id = get_current_ray_worker_id()
    with memray.Tracker(
        f"dedupe_{worker_id}_{task_id}.bin"
    ) if enable_profiler else nullcontext():
        start = time.time()
        logger.info(f"Resolved materialize task obj refs...")
        dedupe_task_indices, obj_refs = zip(*dedupe_task_idx_and_obj_id_tuples)
        # this depends on `ray.get` result order matching input order, as per the
        # contract established in: https://github.com/ray-project/ray/pull/16763
        src_file_records_list = object_store.get_many(list(obj_refs))
        all_src_file_records = defaultdict(list)
        for i, src_file_records in enumerate(src_file_records_list):
            dedupe_task_idx = dedupe_task_indices[i]
            for src_dfl, record_numbers in src_file_records.items():
                all_src_file_records[src_dfl].append(
                    (record_numbers, repeat(dedupe_task_idx, len(record_numbers)))
                )
        manifest_cache = {}
        materialized_results: List[MaterializeResult] = []
        record_batch_tables = RecordBatchTables(max_records_per_output_file)
        count_of_src_dfl = 0
        manifest_entry_list_reference = []
        referenced_pyarrow_write_results = []
        for src_dfl in sorted(all_src_file_records.keys()):
            record_numbers_dd_task_idx_tpl_list: List[
                Tuple[DeltaFileLocatorToRecords, repeat]
            ] = all_src_file_records[src_dfl]
            record_numbers_tpl, dedupe_task_idx_iter_tpl = zip(
                *record_numbers_dd_task_idx_tpl_list
            )
            is_src_partition_file_np = src_dfl.is_source_delta
            src_stream_position_np = src_dfl.stream_position
            src_file_idx_np = src_dfl.file_index
            src_file_record_count = src_dfl.file_record_count.item()
            count_of_src_dfl += 1
            src_file_partition_locator = (
                source_partition_locator
                if is_src_partition_file_np
                else round_completion_info.compacted_delta_locator.partition_locator
            )
            delta_locator = DeltaLocator.of(
                src_file_partition_locator,
                src_stream_position_np.item(),
            )
            dl_digest = delta_locator.digest()
            manifest = manifest_cache.setdefault(
                dl_digest,
                deltacat_storage.get_delta_manifest(
                    delta_locator, **deltacat_storage_kwargs
                ),
            )

            if read_kwargs_provider is None:
                # for delimited text output, disable type inference to prevent
                # unintentional type-casting side-effects and improve performance
                if compacted_file_content_type in DELIMITED_TEXT_CONTENT_TYPES:
                    read_kwargs_provider = ReadKwargsProviderPyArrowCsvPureUtf8()
                # enforce a consistent schema if provided, when reading files into PyArrow tables
                elif schema is not None:
                    read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(
                        schema=schema
                    )
            record_numbers = chain.from_iterable(record_numbers_tpl)
            record_numbers_length = 0
            mask_pylist = list(repeat(False, src_file_record_count))
            for record_number in record_numbers:
                record_numbers_length += 1
                mask_pylist[record_number] = True
            if (
                round_completion_info
                and enable_manifest_entry_copy_by_reference
                and record_numbers_length == src_file_record_count
                and src_file_partition_locator
                == round_completion_info.compacted_delta_locator.partition_locator
            ):
                logger.debug(
                    f"Untouched manifest file found, "
                    f"record numbers length: {record_numbers_length} "
                    f"same as downloaded table length: {src_file_record_count}"
                )
                untouched_src_manifest_entry = manifest.entries[src_file_idx_np.item()]
                manifest_entry_list_reference.append(untouched_src_manifest_entry)
                referenced_pyarrow_write_result = PyArrowWriteResult.of(
                    1,
                    untouched_src_manifest_entry.meta.source_content_length,
                    untouched_src_manifest_entry.meta.content_length,
                    src_file_record_count,
                )
                referenced_pyarrow_write_results.append(referenced_pyarrow_write_result)
            else:
                pa_table, download_delta_manifest_entry_time = timed_invocation(
                    deltacat_storage.download_delta_manifest_entry,
                    Delta.of(delta_locator, None, None, None, manifest),
                    src_file_idx_np.item(),
                    file_reader_kwargs_provider=read_kwargs_provider,
                    **deltacat_storage_kwargs,
                )
                logger.debug(
                    f"Time taken for materialize task"
                    f" to download delta locator {delta_locator} with entry ID {src_file_idx_np.item()}"
                    f" is: {download_delta_manifest_entry_time}s"
                )
                mask = pa.array(mask_pylist)
                pa_table = pa_table.filter(mask)
                record_batch_tables.append(pa_table)
                if record_batch_tables.has_batches():
                    batched_tables = record_batch_tables.evict()
                    materialized_results.append(_materialize(batched_tables))

        if record_batch_tables.has_remaining():
            materialized_results.append(_materialize(record_batch_tables.remaining))

        logger.info(f"Got {count_of_src_dfl} source delta files during materialize")
        referenced_manifest_delta = (
            _stage_delta_from_manifest_entry_reference_list(
                manifest_entry_list_reference, partition
            )
            if manifest_entry_list_reference
            else None
        )

        merged_materialized_delta = [mr.delta for mr in materialized_results]
        merged_materialized_delta.append(referenced_manifest_delta)
        merged_delta = Delta.merge_deltas(
            [d for d in merged_materialized_delta if d is not None]
        )

        write_results_union = [*referenced_pyarrow_write_results]
        if materialized_results:
            for mr in materialized_results:
                write_results_union.append(mr.pyarrow_write_result)
        write_result = PyArrowWriteResult.union(write_results_union)
        referenced_write_result = PyArrowWriteResult.union(
            referenced_pyarrow_write_results
        )

        if referenced_manifest_delta:
            logger.info(
                f"Got delta with {len(referenced_manifest_delta.manifest.entries)} referenced manifest entries"
            )
            assert referenced_write_result.files == len(
                referenced_manifest_delta.manifest.entries
            ), "The files referenced must match with the entries in the delta"

        assert write_result.files == len(
            merged_delta.manifest.entries
        ), "The total number of files written by materialize must match manifest entries"

        logger.debug(
            f"{write_result.files} files written"
            f" with records: {write_result.records}"
        )

        logger.info(f"Finished materialize task...")
        end = time.time()
        duration = end - start

        emit_metrics_time = 0.0
        if metrics_config:
            emit_result, latency = timed_invocation(
                func=emit_timer_metrics,
                metrics_name="materialize",
                value=duration,
                metrics_config=metrics_config,
            )
            emit_metrics_time = latency
        logger.info(f"Materialize task ended in {end - start}s")

        peak_memory_usage_bytes = get_current_node_peak_memory_usage_in_bytes()

        # Merge all new deltas into one for this materialize bucket index
        merged_materialize_result = MaterializeResult.of(
            merged_delta,
            mat_bucket_index,
            write_result,
            referenced_write_result,
            np.double(peak_memory_usage_bytes),
            np.double(emit_metrics_time),
            np.double(time.time()),
        )

        return merged_materialize_result



import unittest
import numpy as np
from unittest import mock


class MockPyMemcacheClient:
    def __init__(self) -> None:
        self.store = {}

    def set_many(self, value_dict, *args, **kwargs):
        self.store.update(value_dict)
        return []

    def get_many(self, keys, *args, **kwargs):
        result = {}
        for key in keys:
            result[key] = self.store[key]

        return result

    def set(self, key, value, *args, **kwargs):
        self.store[key] = value
        return True

    def get(self, key, *args, **kwargs):
        return self.store.get(key)


class TestMemcachedObjectStore(unittest.TestCase):

    TEST_VALUE_LARGE = "test-value-greater-than-10-bytes"

    def setUp(self):
        from deltacat.io.memcached_object_store import MemcachedObjectStore

        self.object_store = MemcachedObjectStore(
            storage_node_ips=["172.1.1.1", "172.2.2.2", "172.3.3.3"],
            max_item_size_bytes=10,
        )

    @mock.patch("deltacat.io.memcached_object_store.Client")
    @mock.patch("deltacat.io.memcached_object_store.RetryingClient")
    def test_put_many_sanity(
        self,
        mock_retrying_client,
        mock_client,
    ):
        mock_client.return_value = MockPyMemcacheClient()
        mock_retrying_client.return_value = mock_client.return_value

        result = self.object_store.put_many(["a", "b", "c"])

        # assert
        self.assertEqual(3, len(result))
        self.assertRegex(result[0], ".*_.*_.*")
        expected = self.object_store.get_many(result)
        self.assertEqual(expected, ["a", "b", "c"])

    @mock.patch("deltacat.io.memcached_object_store.Client")
    @mock.patch("deltacat.io.memcached_object_store.RetryingClient")
    def test_put_many_greater_than_max_item_size(
        self,
        mock_retrying_client,
        mock_client,
    ):
        mock_client.return_value = MockPyMemcacheClient()
        mock_retrying_client.return_value = mock_client.return_value

        put_request = [np.arange(99), np.zeros(1000), np.zeros(100)]
        result = self.object_store.put_many(put_request)

        # assert
        self.assertEqual(3, len(result))
        self.assertRegex(result[0], ".*_.*_.*")
        expected = self.object_store.get_many(result)
        for index, val in enumerate(expected):
            self.assertTrue(np.array_equal(val, put_request[index]))

    @mock.patch("deltacat.io.memcached_object_store.Client")
    @mock.patch("deltacat.io.memcached_object_store.RetryingClient")
    def test_put_many_when_cache_fails(self, mock_retrying_client, mock_client):
        mock_retrying_client.return_value = mock_client.return_value
        mock_client.return_value.set_many.return_value = ["abcd"]

        with self.assertRaises(RuntimeError):
            self.object_store.put_many(["a", "b"])

        self.assertEqual(1, mock_client.return_value.set_many.call_count)

    @mock.patch("deltacat.io.memcached_object_store.Client")
    @mock.patch("deltacat.io.memcached_object_store.RetryingClient")
    def test_get_many_sanity(self, mock_retrying_client, mock_client):
        mock_client.return_value = MockPyMemcacheClient()
        mock_retrying_client.return_value = mock_client.return_value

        # setup
        ref1 = self.object_store.put("a")
        ref2 = self.object_store.put(np.arange(100))

        # action
        result = self.object_store.get_many([ref2, ref1])

        # assert
        self.assertEqual(2, len(result))
        self.assertEqual(result[1], "a")
        self.assertTrue(
            np.array_equal(result[0], np.arange(100)), f"Arrays not equal: {result[0]}"
        )

    @mock.patch("deltacat.io.memcached_object_store.Client")
    @mock.patch("deltacat.io.memcached_object_store.RetryingClient")
    def test_get_many_when_cache_expired(self, mock_retrying_client, mock_client):
        mock_client.return_value.get_many.return_value = {"uuid_ip_1": "value1"}
        mock_retrying_client.return_value = mock_client.return_value

        with self.assertRaises(AssertionError):
            self.object_store.get_many(["uuid_ip_1", "uuid2_ip_1"])

        self.assertEqual(1, mock_client.return_value.get_many.call_count)

    @mock.patch("deltacat.io.memcached_object_store.Client")
    @mock.patch("deltacat.io.memcached_object_store.RetryingClient")
    def test_get_and_put_sanity(self, mock_retrying_client, mock_client):
        # setup
        mock_client.return_value = MockPyMemcacheClient()
        mock_retrying_client.return_value = mock_client.return_value

        # action
        ref = self.object_store.put(self.TEST_VALUE_LARGE)
        result = self.object_store.get(ref)

        # assert
        self.assertEqual(self.TEST_VALUE_LARGE, result)
        _, _, count = ref.split(self.object_store.SEPARATOR)
        self.assertTrue(int(count) > 1, f"The chunks={count} is not > 1")

    @mock.patch("deltacat.io.memcached_object_store.Client")
    @mock.patch("deltacat.io.memcached_object_store.RetryingClient")
    def test_get_when_cache_fails(self, mock_retrying_client, mock_client):
        mock_client.return_value.get.side_effect = RuntimeError()
        mock_retrying_client.return_value = mock_client.return_value

        with self.assertRaises(RuntimeError):
            self.object_store.get("test_ip_1")

        self.assertEqual(1, mock_client.return_value.get.call_count)

    @mock.patch("deltacat.io.memcached_object_store.Client")
    @mock.patch("deltacat.io.memcached_object_store.RetryingClient")
    def test_put_single_chunk(self, mock_retrying_client, mock_client):
        from deltacat.io.memcached_object_store import MemcachedObjectStore

        object_store = MemcachedObjectStore(storage_node_ips=["1.2.2.2"])
        mock_retrying_client.return_value = mock_client.return_value
        mock_client.return_value.set.return_value = True

        result = object_store.put("test")

        self.assertIsNotNone(result)
        self.assertEqual(1, mock_client.return_value.set.call_count)

    @mock.patch("deltacat.io.memcached_object_store.Client")
    @mock.patch("deltacat.io.memcached_object_store.RetryingClient")
    def test_put_when_cache_fails(self, mock_retrying_client, mock_client):
        mock_client.return_value.set.return_value = False
        mock_retrying_client.return_value = mock_client.return_value

        with self.assertRaises(RuntimeError):
            self.object_store.put("test_ip")

        self.assertEqual(1, mock_client.return_value.set.call_count)

    @mock.patch("deltacat.io.memcached_object_store.Client")
    @mock.patch("deltacat.io.memcached_object_store.RetryingClient")
    @mock.patch("deltacat.io.memcached_object_store.socket")
    def test_put_when_storage_nodes_none(
        self, mock_socket, mock_retrying_client, mock_client
    ):
        from deltacat.io.memcached_object_store import MemcachedObjectStore

        object_store = MemcachedObjectStore()

        # setup
        mock_socket.gethostbyname.return_value = "0.0.0.0"
        mock_socket.gethostname.return_value = "localhost"
        mock_client.return_value = MockPyMemcacheClient()
        mock_retrying_client.return_value = mock_client.return_value

        # action
        ref = object_store.put(self.TEST_VALUE_LARGE)

        # assert
        result = self.object_store.get(ref)
        self.assertEqual(result, self.TEST_VALUE_LARGE)



import json

import unittest


class TestCompactPartitionParams(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        from deltacat.types.media import ContentType
        from deltacat.utils.metrics import MetricsConfig, MetricsTarget

        cls.VALID_COMPACT_PARTITION_PARAMS = {
            "compaction_artifact_s3_bucket": "foobar",
            "compacted_file_content_type": ContentType.PARQUET,
            "deltacat_storage": "foobar",
            "destination_partition_locator": {
                "streamLocator": {
                    "tableVersionLocator": {
                        "tableLocator": {
                            "namespaceLocator": {"namespace": "testNamespaceLocator"},
                            "tableName": "TABLE_FOO",
                        },
                        "tableVersion": "1",
                    },
                    "streamId": "foobar",
                    "storageType": "fooType",
                },
                "partitionValues": [],
                "partitionId": None,
            },
            "hash_bucket_count": 200,
            "last_stream_position_to_compact": 168000000000,
            "list_deltas_kwargs": {"equivalent_table_types": []},
            "primary_keys": {"id"},
            "properties": {
                "parent_stream_position": "1688000000000",
            },
            "read_kwargs_provider": "foo",
            "rebase_source_partition_high_watermark": 1688000000000,
            "rebase_source_partition_locator": {
                "stream_locator": {
                    "table_version_locator": {
                        "table_locator": {
                            "namespace_locator": {"namespace": "testNamespaceLocator"},
                            "tableName": "TABLE_FOO",
                        },
                        "table_version": "1",
                    },
                    "streamId": "foobar",
                    "storageType": "fooType",
                },
                "partitionValues": [],
                "partitionId": "79612ea39ac5493eae925abe60767d42",
            },
            "s3_table_writer_kwargs": {
                "version": "1.0",
                "flavor": "foobar",
                "coerce_timestamps": "ms",
            },
            "source_partition_locator": {
                "streamLocator": {
                    "tableVersionLocator": {
                        "tableLocator": {
                            "namespaceLocator": {"namespace": "testNamespaceLocator"},
                            "tableName": "TABLE_FOO",
                        },
                        "tableVersion": "2",
                    },
                    "streamId": "foobar",
                    "storageType": "fooType",
                },
                "partitionValues": [],
                "partitionId": "79612ea39ac5493eae925abe60767d42",
            },
            "metrics_config": MetricsConfig("us-east-1", MetricsTarget.CLOUDWATCH_EMF),
        }

        super().setUpClass()

    def test_serialize_returns_json_string(self):
        from deltacat.compute.compactor.model.compact_partition_params import (
            CompactPartitionParams,
        )

        params = CompactPartitionParams.of(
            TestCompactPartitionParams.VALID_COMPACT_PARTITION_PARAMS
        )
        serialized_params = params.serialize()
        assert isinstance(serialized_params, str)

    def test_serialize_returns_json_string_with_all_fields(self):
        from deltacat.compute.compactor.model.compact_partition_params import (
            CompactPartitionParams,
        )

        params = CompactPartitionParams.of(
            TestCompactPartitionParams.VALID_COMPACT_PARTITION_PARAMS
        )
        serialized_params = params.serialize()
        assert isinstance(serialized_params, str)
        assert (
            json.loads(serialized_params)["compacted_file_content_type"]
            == params.compacted_file_content_type
        )
        assert (
            json.loads(serialized_params)["compaction_artifact_s3_bucket"]
            == params.compaction_artifact_s3_bucket
        )
        assert (
            json.loads(serialized_params)["hash_bucket_count"]
            == params.hash_bucket_count
        )
        assert (
            json.loads(serialized_params)["last_stream_position_to_compact"]
            == params.last_stream_position_to_compact
        )
        assert (
            json.loads(serialized_params)["list_deltas_kwargs"]
            == params.list_deltas_kwargs
        )
        assert json.loads(serialized_params)["primary_keys"] == params.primary_keys
        assert (
            json.loads(serialized_params)["rebase_source_partition_high_watermark"]
            == params.rebase_source_partition_high_watermark
        )
        assert (
            json.loads(serialized_params)["rebase_source_partition_locator"]
            == params.rebase_source_partition_locator
        )
        assert (
            json.loads(serialized_params)["source_partition_locator"]
            == params.source_partition_locator
        )
        assert (
            json.loads(serialized_params)["destination_partition_locator"]
            == params.destination_partition_locator
        )
        assert (
            json.loads(serialized_params)["metrics_config"]["metrics_target"]
            == params.metrics_config.metrics_target
        )

    def test_serialize_handles_sets(self):
        from deltacat.compute.compactor.model.compact_partition_params import (
            CompactPartitionParams,
        )

        params = CompactPartitionParams.of(
            {
                **TestCompactPartitionParams.VALID_COMPACT_PARTITION_PARAMS,
                "primary_keys": {"foo", "bar", "baz"},
            }
        )
        serialized_params = params.serialize()
        self.assertCountEqual(
            json.loads(serialized_params)["primary_keys"], ["foo", "bar", "baz"]
        )

    def test_serialize_handles_objects_with_toJSON_method(self):
        from deltacat.compute.compactor.model.compact_partition_params import (
            CompactPartitionParams,
        )

        class MyObject:
            def toJSON(self) -> str:
                return "my-json-object"

        params = CompactPartitionParams.of(
            {
                **TestCompactPartitionParams.VALID_COMPACT_PARTITION_PARAMS,
                "compacted_file_content_type": MyObject(),
            }
        )
        serialized_params = params.serialize()
        assert (
            json.loads(serialized_params)["compacted_file_content_type"]
            == "my-json-object"
        )

    def test_json_handler_for_compact_partition_params_serializes_set_to_list(self):
        from deltacat.compute.compactor.model.compact_partition_params import (
            CompactPartitionParams,
        )

        my_set = {1, 2, 3}
        json_string = json.dumps(
            my_set,
            default=CompactPartitionParams.json_handler_for_compact_partition_params,
        )
        assert json.loads(json_string) == [1, 2, 3]

    def test_json_handler_for_compact_partition_params_serializes_object_with_toJSON_method_to_dict(
        self,
    ):
        from dataclasses import dataclass

        from deltacat.compute.compactor.model.compact_partition_params import (
            CompactPartitionParams,
        )

        @dataclass
        class DummyObject:
            some_property: str = "foo"

            def toJSON(self):
                return self.__dict__

        dummy_object = DummyObject()
        json_string = json.dumps(
            dummy_object,
            default=CompactPartitionParams.json_handler_for_compact_partition_params,
        )
        assert json.loads(json_string) == {"some_property": "foo"}

    def test_json_handler_for_compact_partition_params_returns_class_name_for_unknown_objects(
        self,
    ):
        from deltacat.compute.compactor.model.compact_partition_params import (
            CompactPartitionParams,
        )

        my_object = object()
        json_string = json.dumps(
            my_object,
            default=CompactPartitionParams.json_handler_for_compact_partition_params,
        )
        assert json.loads(json_string) == "object"



from dataclasses import dataclass

import logging

from aws_embedded_metrics import metric_scope
from aws_embedded_metrics.config import get_config
from aws_embedded_metrics.logger.metrics_logger import MetricsLogger
from deltacat import logs
from enum import Enum
from typing import Dict, Any, List, Callable
from deltacat.aws.clients import resource_cache
from datetime import datetime

from ray.util import get_node_ip_address

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))

DEFAULT_DELTACAT_METRICS_NAMESPACE = "ray-deltacat-metrics"
DEFAULT_DELTACAT_LOG_GROUP_NAME = "ray-deltacat-metrics-EMF-logs"
DEFAULT_DELTACAT_LOG_STREAM_CALLABLE = get_node_ip_address


class MetricsTarget(str, Enum):
    CLOUDWATCH = "cloudwatch"
    CLOUDWATCH_EMF = "cloudwatch_emf"


@dataclass
class MetricsConfig:
    def __init__(
        self,
        region: str,
        metrics_target: MetricsTarget,
        metrics_namespace: str = DEFAULT_DELTACAT_METRICS_NAMESPACE,
        metrics_dimensions: List[Dict[str, str]] = [],
        metrics_kwargs: Dict[str, Any] = {},
    ):
        self.region = region
        self.metrics_target = metrics_target
        self.metrics_namespace = metrics_namespace
        self.metrics_dimensions = metrics_dimensions
        self.metrics_kwargs = metrics_kwargs


class MetricsType(str, Enum):
    TIMER = "timer"


def _build_metrics_name(metrics_type: Enum, metrics_name: str) -> str:
    metrics_name_with_type = f"{metrics_name}_{metrics_type}"
    return metrics_name_with_type


def _build_cloudwatch_metrics(
    metrics_name: str,
    metrics_type: Enum,
    value: str,
    metrics_dimensions: List[Dict[str, str]],
    timestamp: datetime,
    **kwargs,
) -> Dict[str, Any]:
    metrics_name_with_type = _build_metrics_name(metrics_type, metrics_name)
    return [
        {
            "MetricName": f"{metrics_name_with_type}",
            "Dimensions": metrics_dimensions,
            "Timestamp": timestamp,
            "Value": value,
            **kwargs,
        }
    ]


def _emit_metrics(
    metrics_name: str,
    metrics_type: Enum,
    metrics_config: MetricsConfig,
    value: str,
    **kwargs,
) -> None:
    metrics_target = metrics_config.metrics_target
    assert isinstance(
        metrics_target, MetricsTarget
    ), f"{metrics_target} is not a valid supported metrics target type! "
    if metrics_target in METRICS_TARGET_TO_EMITTER_DICT:
        METRICS_TARGET_TO_EMITTER_DICT.get(metrics_target)(
            metrics_name=metrics_name,
            metrics_type=metrics_type,
            metrics_config=metrics_config,
            value=value,
            **kwargs,
        )
    else:
        logger.warning(f"{metrics_target} is not a supported metrics target type.")


def _emit_cloudwatch_metrics(
    metrics_name: str,
    metrics_type: Enum,
    metrics_config: MetricsConfig,
    value: str,
    **kwargs,
) -> None:
    current_time = datetime.now()
    current_instance_region = metrics_config.region
    cloudwatch_resource = resource_cache("cloudwatch", current_instance_region)
    cloudwatch_client = cloudwatch_resource.meta.client

    metrics_namespace = metrics_config.metrics_namespace
    metrics_dimensions = metrics_config.metrics_dimensions

    metrics_data = _build_cloudwatch_metrics(
        metrics_name,
        metrics_type,
        value,
        metrics_dimensions,
        current_time,
        **metrics_config.metrics_kwargs,
        **kwargs,
    )

    response = None
    try:
        response = cloudwatch_client.put_metric_data(
            Namespace=metrics_namespace,
            MetricData=metrics_data,
        )
    except Exception as e:
        logger.warning(
            f"Failed to publish Cloudwatch metrics with name: {metrics_name}, "
            f"type: {metrics_type}, with exception: {e}, response: {response}"
        )


@metric_scope
def _emit_cloudwatch_emf_metrics(
    metrics_name: str,
    metrics_type: Enum,
    metrics_config: MetricsConfig,
    value: str,
    metrics: MetricsLogger,
    **kwargs,
) -> None:
    """
    Publishes CloudWatch EMF logs, which are metrics embedded in CloudWatch logs.

    There are multiple possible metrics_kwargs keys that can be expected in this method.
    The documentation can be found here: github.com/awslabs/aws-embedded-metrics-python

    Keyword arguments currently being checked for:
     - log_group_name (str): specifies the CloudWatch log group name
     - log_stream_name (callable): specifies the log stream name to be used within the log group
    """
    current_time = datetime.now()
    dimensions = dict(
        [(dim["Name"], dim["Value"]) for dim in metrics_config.metrics_dimensions]
    )
    metrics_name_with_type = _build_metrics_name(metrics_type, metrics_name)
    try:
        metrics.set_timestamp(current_time)
        metrics.set_dimensions(dimensions)
        metrics.set_namespace(metrics_config.metrics_namespace)

        metrics_kwargs = metrics_config.metrics_kwargs

        emf_config = get_config()
        emf_config.log_group_name = (
            metrics_kwargs["log_group_name"]
            if "log_group_name" in metrics_kwargs
            else DEFAULT_DELTACAT_LOG_GROUP_NAME
        )

        # use a callable to differentiate many possible log streams
        emf_config.log_stream_name = (
            metrics_kwargs["log_stream_name"]()
            if "log_stream_name" in metrics_kwargs
            else DEFAULT_DELTACAT_LOG_STREAM_CALLABLE()
        )

        metrics.put_metric(metrics_name_with_type, value)
    except Exception as e:
        logger.warning(
            f"Failed to publish Cloudwatch EMF metrics with name: {metrics_name_with_type}, "
            f"type: {metrics_type}, with exception: {e}"
        )


METRICS_TARGET_TO_EMITTER_DICT: Dict[str, Callable] = {
    MetricsTarget.CLOUDWATCH: _emit_cloudwatch_metrics,
    MetricsTarget.CLOUDWATCH_EMF: _emit_cloudwatch_emf_metrics,
}


def emit_timer_metrics(metrics_name, value, metrics_config, **kwargs):
    _emit_metrics(
        metrics_name=metrics_name,
        metrics_type=MetricsType.TIMER,
        metrics_config=metrics_config,
        value=value,
        **kwargs,
    )



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations
from typing import Optional
import pyarrow as pa
import logging
from deltacat import logs
from typing import List, Union
from deltacat.compute.compactor.model.hash_bucket_result import HashBucketResult
from deltacat.compute.compactor.model.dedupe_result import DedupeResult
from deltacat.compute.compactor.model.materialize_result import MaterializeResult
from deltacat.utils.performance import timed_invocation
from deltacat.utils.resources import ClusterUtilization, get_size_of_object_in_bytes
from deltacat.compute.compactor import PyArrowWriteResult

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


class CompactionSessionAuditInfo(dict):

    DEDUPE_STEP_NAME = "dedupe"
    MATERIALIZE_STEP_NAME = "materialize"
    HASH_BUCKET_STEP_NAME = "hashBucket"
    MERGE_STEP_NAME = "merge"

    def __init__(
        self,
        deltacat_version: Optional[str] = None,
        ray_version: Optional[str] = None,
        audit_url: Optional[str] = None,
        **kwargs,
    ):
        self.set_deltacat_version(deltacat_version)
        self.set_ray_version(ray_version)
        self.set_audit_url(audit_url)
        if kwargs:
            self.update(kwargs)

    @property
    def audit_url(self) -> str:
        return self.get("auditUrl")

    @property
    def deltacat_version(self) -> str:
        """
        The deltacat version used to run compaction job.
        """
        return self.get("deltacatVersion")

    @property
    def ray_version(self) -> str:
        """
        The ray distribution version used to run compaction job.
        """
        return self.get("rayVersion")

    @property
    def input_records(self) -> int:
        """
        The total number of records from input deltas that needs to be compacted
        (before deduplication).
        """
        return self.get("inputRecords")

    @property
    def input_file_count(self) -> int:
        """
        The total number of input files that needs to be compacted.
        """
        return self.get("inputFileCount")

    @property
    def uniform_deltas_created(self) -> int:
        """
        The total number of uniform deltas fed into the hash bucket step.
        """
        return self.get("uniformDeltasCreated")

    @property
    def records_deduped(self) -> int:
        """
        The total number of records that were deduplicated. For example,
        if there are 100 records with a particular primary key, 99 records
        will be deduplicated.
        """
        return self.get("recordsDeduped")

    @property
    def input_size_bytes(self) -> float:
        """
        The on-disk size in bytes of the input. Analogous to bytes scanned
        """
        return self.get("inputSizeBytes")

    @property
    def hash_bucket_count(self) -> int:
        """
        Total number of hash buckets used during compaction.
        """
        return self.get("hashBucketCount")

    @property
    def cluster_cpu_max(self) -> float:
        """
        Total cluster cpu allocated for the compaction job. If it is autoscaling cluster,
        max cpu at any time will be reported.
        """
        return self.get("clusterCpuMax")

    @property
    def compaction_time_in_seconds(self) -> float:
        """
        The total time taken by the compaction session to complete.
        """
        return self.get("compactionTimeInSeconds")

    @property
    def total_object_store_memory_used_bytes(self) -> float:
        """
        The total object store memory used by the compaction session across all
        nodes in the entire cluster.
        """
        return self.get("totalObjectStoreMemoryUsedBytes")

    @property
    def peak_memory_used_bytes_per_task(self) -> float:
        """
        The peak memory used by a single process in the compaction job. Note that
        Ray creates a single process to run each hash bucketing, dedupe and
        materialize task and the process is reused. Hence, you may see
        monotonically increasing values. Peak memory is important because,
        the cluster must be scaled to handle  the peak memory per node even
        though average memory usage is low.
        """
        return self.get("peakMemoryUsedBytesPerTask")

    @property
    def peak_memory_used_bytes_per_hash_bucket_task(self) -> float:
        """
        The peak memory used by a single hash bucketing process. For example,
        if peak usage of hash bucketing process is 40GB, it is not safe to run
        more than 3 hash bucketing tasks on a node with 120GB to avoid crashing
        due to memory overflow.
        """
        return self.get("hashBucketTaskPeakMemoryUsedBytes")

    @property
    def peak_memory_used_bytes_per_dedupe_task(self) -> float:
        """
        The peak memory used by a single dedupe python process. Note that
        results may be max of dedupe and hash bucketing as processes are
        reused by Ray to run dedupe and hash bucketing.
        """
        return self.get("dedupeTaskPeakMemoryUsedBytes")

    @property
    def peak_memory_used_bytes_per_materialize_task(self) -> float:
        """
        The peak memory used by a single materialize python process. Note
        that results may be max of materialize, dedupe and hash bucketing as
        processes are reused by Ray to run all compaction steps.
        """
        return self.get("materializeTaskPeakMemoryUsedBytes")

    @property
    def peak_memory_used_bytes_per_merge_task(self) -> float:
        """
        The peak memory used by a single merge python process. Note
        that results may be max of merge, and hash bucketing as
        processes are reused by Ray to run all compaction steps.
        """
        return self.get("mergeTaskPeakMemoryUsedBytes")

    @property
    def hash_bucket_post_object_store_memory_used_bytes(self) -> float:
        """
        The total object store memory used by hash bucketing step across
        cluster, before dedupe is run.
        """
        return self.get("hashBucketPostObjectStoreMemoryUsedBytes")

    @property
    def dedupe_post_object_store_memory_used_bytes(self) -> float:
        """
        The total object store memory used after dedupe step before materialize is run.
        """
        return self.get("dedupePostObjectStoreMemoryUsedBytes")

    @property
    def materialize_post_object_store_memory_used_bytes(self) -> float:
        """
        The total object store memory used after materialize step.
        """
        return self.get("materializePostObjectStoreMemoryUsedBytes")

    @property
    def merge_post_object_store_memory_used_bytes(self) -> float:
        """
        The total object store memory used after merge step.
        """
        return self.get("mergePostObjectStoreMemoryUsedBytes")

    @property
    def materialize_buckets(self) -> int:
        """
        The total number of materialize buckets created.
        """
        return self.get("materializeBuckets")

    @property
    def hash_bucket_time_in_seconds(self) -> float:
        """
        The time taken by hash bucketing step. This includes all hash bucket tasks.
        This includes invoke time.
        """
        return self.get("hashBucketTimeInSeconds")

    @property
    def hash_bucket_invoke_time_in_seconds(self) -> float:
        """
        The time taken to invoke and create all hash bucketing tasks.
        """
        return self.get("hashBucketInvokeTimeInSeconds")

    @property
    def hash_bucket_result_wait_time_in_seconds(self) -> float:
        """
        The time it takes ray.get() to resolve after the last hash bucket task has completed.
        This value may not be accurate at less than 1 second precision.
        """
        return self.get("hashBucketResultWaitTimeInSeconds")

    @property
    def dedupe_time_in_seconds(self) -> float:
        """
        The time taken by dedupe step. This include all dedupe tasks.
        """
        return self.get("dedupeTimeInSeconds")

    @property
    def dedupe_invoke_time_in_seconds(self) -> float:
        """
        The time taken to invoke all dedupe tasks.
        """
        return self.get("dedupeInvokeTimeInSeconds")

    @property
    def dedupe_result_wait_time_in_seconds(self) -> float:
        """
        The time it takes ray.get() to resolve after the last dedupe task has completed.
        This value may not be accurate at less than 1 second precision.
        """
        return self.get("dedupeResultWaitTimeInSeconds")

    @property
    def materialize_time_in_seconds(self) -> float:
        """
        The time taken by materialize step. This includes all materialize tasks.
        """
        return self.get("materializeTimeInSeconds")

    @property
    def materialize_invoke_time_in_seconds(self) -> float:
        """
        The time taken to invoke all materialize tasks.
        """
        return self.get("materializeInvokeTimeInSeconds")

    @property
    def materialize_result_wait_time_in_seconds(self) -> float:
        """
        The time it takes ray.get() to resolve after the last materialize task has completed.
        This value may not be accurate at less than 1 second precision.
        """
        return self.get("materializeResultWaitTimeInSeconds")

    @property
    def merge_result_wait_time_in_seconds(self) -> float:
        """
        The time it takes ray.get() to resolve after the last task has completed.
        This value may not be accurate at less than 1 second precision.
        """
        return self.get("mergeResultWaitTimeInSeconds")

    @property
    def merge_time_in_seconds(self) -> float:
        """
        The time taken by merge step. This includes all merge tasks.
        """
        return self.get("mergeTimeInSeconds")

    @property
    def merge_invoke_time_in_seconds(self) -> float:
        """
        The time taken to invoke all merge tasks.
        """
        return self.get("mergeInvokeTimeInSeconds")

    @property
    def delta_discovery_time_in_seconds(self) -> float:
        """
        The time taken by delta discovery step which is mostly run before hash bucketing is started.
        """
        return self.get("deltaDiscoveryTimeInSeconds")

    @property
    def output_file_count(self) -> int:
        """
        The total number of files in the compacted output (includes untouched files).
        """
        return self.get("outputFileCount")

    @property
    def output_size_bytes(self) -> float:
        """
        The on-disk size of the compacted output including any untouched files.
        """
        return self.get("outputSizeBytes")

    @property
    def output_size_pyarrow_bytes(self) -> float:
        """
        The pyarrow in-memory size of compacted output including any untouched files.
        """
        return self.get("outputSizePyarrowBytes")

    @property
    def total_cluster_memory_bytes(self) -> float:
        """
        The total memory allocated to the cluster.
        """
        return self.get("totalClusterMemoryBytes")

    @property
    def total_cluster_object_store_memory_bytes(self) -> float:
        """
        The total object store memory allocated to the cluster.
        """
        return self.get("totalClusterObjectStoreMemoryBytes")

    @property
    def untouched_file_count(self) -> int:
        """
        The total number of files that were untouched by materialize step.
        """
        return self.get("untouchedFileCount")

    @property
    def untouched_file_ratio(self) -> float:
        """
        The ratio between total number of files untouched and total number of files in the compacted output.
        """
        return self.get("untouchedFileRatio")

    @property
    def untouched_record_count(self) -> int:
        """
        The total number of records untouched during materialization.
        """
        return self.get("untouchedRecordCount")

    @property
    def untouched_size_bytes(self) -> float:
        """
        The on-disk size of the data untouched during materialization.
        """
        return self.get("untouchedSizeBytes")

    @property
    def telemetry_time_in_seconds(self) -> float:
        """
        The total time taken by all the telemetry activity across the nodes in the cluster. This includes
        collecting cluster resources information, emitting metrics, etc.
        """
        return self.get("telemetryTimeInSeconds")

    @property
    def hash_bucket_result_size_bytes(self) -> float:
        """
        The size of the results returned by hash bucket step.
        """
        return self.get("hashBucketResultSize")

    @property
    def dedupe_result_size_bytes(self) -> float:
        """
        The size of the results returned by dedupe step.
        """
        return self.get("dedupeResultSize")

    @property
    def materialize_result_size(self) -> float:
        """
        The size of the results returned by materialize step.
        """
        return self.get("materializeResultSize")

    @property
    def merge_result_size(self) -> float:
        """
        The size of the results returned by merge step.
        """
        return self.get("mergeResultSize")

    @property
    def peak_memory_used_bytes_by_compaction_session_process(self) -> float:
        """
        The peak memory used by the entrypoint for compaction_session.
        """
        return self.get("peakMemoryUsedBytesCompactionSessionProcess")

    @property
    def estimated_in_memory_size_bytes_during_discovery(self) -> float:
        """
        The estimated in-memory size during the discovery. This can be used
        to determine the accuracy of memory estimation logic.
        """
        return self.get("estimatedInMemorySizeBytesDuringDiscovery")

    @property
    def hash_bucket_processed_size_bytes(self) -> int:
        """
        The total size of the input data processed during hash bucket
        """
        return self.get("hashBucketProcessedSizeBytes")

    @property
    def total_cpu_seconds(self) -> float:
        """
        Total number of vCPUs provisioned in the cluster weighted over time.
        """
        return self.get("totalCPUSeconds")

    @property
    def used_cpu_seconds(self) -> float:
        """
        Total used vCPU in the cluster weighted over time.
        """
        return self.get("usedCPUSeconds")

    @property
    def used_memory_gb_seconds(self) -> float:
        """
        The used memory in the cluster weighted over time. This
        determines opportunities for better memory estimation.
        """
        return self.get("usedMemoryGBSeconds")

    @property
    def total_memory_gb_seconds(self) -> float:
        """
        Total memory in the cluster weighted over time in GB.
        """
        return self.get("totalMemoryGBSeconds")

    @property
    def pyarrow_version(self) -> str:
        """
        The version of PyArrow used.
        """
        return self.get("pyarrowVersion")

    @property
    def compactor_version(self) -> str:
        """
        The version of the compactor used to compact.
        """
        return self.get("compactorVersion")

    # Setters follow

    def set_audit_url(self, audit_url: str) -> CompactionSessionAuditInfo:
        self["auditUrl"] = audit_url
        return self

    def set_deltacat_version(self, version: str) -> CompactionSessionAuditInfo:
        self["deltacatVersion"] = version
        return self

    def set_ray_version(self, version: str) -> CompactionSessionAuditInfo:
        self["rayVersion"] = version
        return self

    def set_input_records(self, input_records: int) -> CompactionSessionAuditInfo:
        self["inputRecords"] = input_records
        return self

    def set_input_file_count(self, input_file_count: int) -> CompactionSessionAuditInfo:
        self["inputFileCount"] = input_file_count
        return self

    def set_uniform_deltas_created(
        self, uniform_deltas_created: int
    ) -> CompactionSessionAuditInfo:
        self["uniformDeltasCreated"] = uniform_deltas_created
        return self

    def set_records_deduped(self, records_deduped: int) -> CompactionSessionAuditInfo:
        self["recordsDeduped"] = records_deduped
        return self

    def set_input_size_bytes(
        self, input_size_bytes: float
    ) -> CompactionSessionAuditInfo:
        self["inputSizeBytes"] = input_size_bytes
        return self

    def set_hash_bucket_count(
        self, hash_bucket_count: int
    ) -> CompactionSessionAuditInfo:
        self["hashBucketCount"] = hash_bucket_count
        return self

    def set_cluster_cpu_max(self, cluster_cpu_max: float) -> CompactionSessionAuditInfo:
        self["clusterCpuMax"] = cluster_cpu_max
        return self

    def set_compaction_time_in_seconds(
        self, compaction_time_in_seconds: float
    ) -> CompactionSessionAuditInfo:
        self["compactionTimeInSeconds"] = compaction_time_in_seconds
        return self

    def set_total_object_store_memory_used_bytes(
        self, total_object_store_memory_used_bytes: float
    ) -> CompactionSessionAuditInfo:
        self["totalObjectStoreMemoryUsedBytes"] = total_object_store_memory_used_bytes
        return self

    def set_peak_memory_used_bytes_per_task(
        self, peak_memory_used_bytes: float
    ) -> CompactionSessionAuditInfo:
        self["peakMemoryUsedBytesPerTask"] = peak_memory_used_bytes
        return self

    def set_peak_memory_used_bytes_per_hash_bucket_task(
        self, peak_memory_used_bytes_per_hash_bucket_task: float
    ) -> CompactionSessionAuditInfo:
        self[
            "hashBucketTaskPeakMemoryUsedBytes"
        ] = peak_memory_used_bytes_per_hash_bucket_task
        return self

    def set_peak_memory_used_bytes_per_dedupe_task(
        self, peak_memory_used_bytes_per_dedupe_task: float
    ) -> CompactionSessionAuditInfo:
        self["dedupeTaskPeakMemoryUsedBytes"] = peak_memory_used_bytes_per_dedupe_task
        return self

    def set_peak_memory_used_bytes_per_materialize_task(
        self, peak_memory_used_bytes_per_materialize_task: float
    ) -> CompactionSessionAuditInfo:
        self[
            "materializeTaskPeakMemoryUsedBytes"
        ] = peak_memory_used_bytes_per_materialize_task
        return self

    def set_peak_memory_used_bytes_per_merge_task(
        self, peak_memory_used_bytes: float
    ) -> CompactionSessionAuditInfo:
        self["mergeTaskPeakMemoryUsedBytes"] = peak_memory_used_bytes
        return self

    def set_hash_bucket_post_object_store_memory_used_bytes(
        self, object_store_memory_used_bytes_by_hb: float
    ) -> CompactionSessionAuditInfo:
        self[
            "hashBucketPostObjectStoreMemoryUsedBytes"
        ] = object_store_memory_used_bytes_by_hb
        return self

    def set_dedupe_post_object_store_memory_used_bytes(
        self, object_store_memory_used_bytes_by_dedupe: float
    ) -> CompactionSessionAuditInfo:
        self[
            "dedupePostObjectStoreMemoryUsedBytes"
        ] = object_store_memory_used_bytes_by_dedupe
        return self

    def set_materialize_post_object_store_memory_used_bytes(
        self, object_store_memory_used_bytes_by_dedupe: float
    ) -> CompactionSessionAuditInfo:
        self[
            "materializePostObjectStoreMemoryUsedBytes"
        ] = object_store_memory_used_bytes_by_dedupe
        return self

    def set_merge_post_object_store_memory_used_bytes(
        self, object_store_memory_used_bytes: float
    ) -> CompactionSessionAuditInfo:
        self["mergePostObjectStoreMemoryUsedBytes"] = object_store_memory_used_bytes
        return self

    def set_materialize_buckets(
        self, materialize_buckets: int
    ) -> CompactionSessionAuditInfo:
        self["materializeBuckets"] = materialize_buckets
        return self

    def set_hash_bucket_time_in_seconds(
        self, hash_bucket_time_in_seconds: float
    ) -> CompactionSessionAuditInfo:
        self["hashBucketTimeInSeconds"] = hash_bucket_time_in_seconds
        return self

    def set_hash_bucket_invoke_time_in_seconds(
        self, hash_bucket_invoke_time: float
    ) -> CompactionSessionAuditInfo:
        self["hashBucketInvokeTimeInSeconds"] = hash_bucket_invoke_time
        return self

    def set_hash_bucket_result_wait_time_in_seconds(
        self, wait_time: float
    ) -> CompactionSessionAuditInfo:
        self.get["hashBucketResultWaitTimeInSeconds"] = wait_time
        return self

    def set_dedupe_time_in_seconds(
        self, dedupe_time_in_seconds: float
    ) -> CompactionSessionAuditInfo:
        self["dedupeTimeInSeconds"] = dedupe_time_in_seconds
        return self

    def set_dedupe_invoke_time_in_seconds(
        self, dedupe_invoke_time: float
    ) -> CompactionSessionAuditInfo:
        self["dedupeInvokeTimeInSeconds"] = dedupe_invoke_time
        return self

    def set_dedupe_result_wait_time_in_seconds(
        self, wait_time: float
    ) -> CompactionSessionAuditInfo:
        self.get["dedupeResultWaitTimeInSeconds"] = wait_time
        return self

    def set_materialize_time_in_seconds(
        self, materialize_time_in_seconds: float
    ) -> CompactionSessionAuditInfo:
        self["materializeTimeInSeconds"] = materialize_time_in_seconds
        return self

    def set_materialize_invoke_time_in_seconds(
        self, materialize_invoke_time: float
    ) -> CompactionSessionAuditInfo:
        self["materializeInvokeTimeInSeconds"] = materialize_invoke_time
        return self

    def set_materialize_result_wait_time_in_seconds(
        self, wait_time: float
    ) -> CompactionSessionAuditInfo:
        self.get["materializeResultWaitTimeInSeconds"] = wait_time
        return self

    def set_merge_time_in_seconds(
        self, time_in_seconds: float
    ) -> CompactionSessionAuditInfo:
        self["mergeTimeInSeconds"] = time_in_seconds
        return self

    def set_merge_invoke_time_in_seconds(
        self, invoke_time: float
    ) -> CompactionSessionAuditInfo:
        self["mergeInvokeTimeInSeconds"] = invoke_time
        return self

    def set_merge_result_wait_time_in_seconds(
        self, wait_time: float
    ) -> CompactionSessionAuditInfo:
        self.get["mergeResultWaitTimeInSeconds"] = wait_time
        return self

    def set_delta_discovery_time_in_seconds(
        self, delta_discovery_time_in_seconds: float
    ) -> CompactionSessionAuditInfo:
        self["deltaDiscoveryTimeInSeconds"] = delta_discovery_time_in_seconds
        return self

    def set_output_file_count(
        self, output_file_count: float
    ) -> CompactionSessionAuditInfo:
        self["outputFileCount"] = output_file_count
        return self

    def set_output_size_bytes(
        self, output_size_bytes: float
    ) -> CompactionSessionAuditInfo:
        self["outputSizeBytes"] = output_size_bytes
        return output_size_bytes

    def set_output_size_pyarrow_bytes(
        self, output_size_pyarrow_bytes: float
    ) -> CompactionSessionAuditInfo:
        self["outputSizePyarrowBytes"] = output_size_pyarrow_bytes
        return output_size_pyarrow_bytes

    def set_total_cluster_memory_bytes(
        self, total_cluster_memory_bytes: float
    ) -> CompactionSessionAuditInfo:
        self["totalClusterMemoryBytes"] = total_cluster_memory_bytes
        return self

    def set_total_cluster_object_store_memory_bytes(
        self, total_cluster_object_store_memory_bytes: float
    ) -> CompactionSessionAuditInfo:
        self[
            "totalClusterObjectStoreMemoryBytes"
        ] = total_cluster_object_store_memory_bytes
        return self

    def set_untouched_file_count(
        self, untouched_file_count: int
    ) -> CompactionSessionAuditInfo:
        self["untouchedFileCount"] = untouched_file_count
        return self

    def set_untouched_file_ratio(
        self, untouched_file_ratio: float
    ) -> CompactionSessionAuditInfo:
        self["untouchedFileRatio"] = untouched_file_ratio
        return self

    def set_untouched_record_count(
        self, untouched_record_count: int
    ) -> CompactionSessionAuditInfo:
        self["untouchedRecordCount"] = untouched_record_count
        return self

    def set_untouched_size_bytes(
        self, untouched_size_bytes: float
    ) -> CompactionSessionAuditInfo:
        self["untouchedSizeBytes"] = untouched_size_bytes
        return self

    def set_telemetry_time_in_seconds(
        self, telemetry_time_in_seconds: float
    ) -> CompactionSessionAuditInfo:
        self["telemetryTimeInSeconds"] = telemetry_time_in_seconds
        return self

    def set_hash_bucket_result_size_bytes(
        self, hash_bucket_result_size_bytes: float
    ) -> CompactionSessionAuditInfo:
        self["hashBucketResultSize"] = hash_bucket_result_size_bytes
        return self

    def set_dedupe_result_size_bytes(
        self, dedupe_result_size_bytes: float
    ) -> CompactionSessionAuditInfo:
        self["dedupeResultSize"] = dedupe_result_size_bytes
        return self

    def set_materialize_result_size_bytes(
        self, materialize_result_size_bytes: float
    ) -> CompactionSessionAuditInfo:
        self["materializeResultSize"] = materialize_result_size_bytes
        return self

    def set_merge_result_size_bytes(
        self, merge_result_size_bytes: float
    ) -> CompactionSessionAuditInfo:
        self["mergeResultSize"] = merge_result_size_bytes
        return self

    def set_peak_memory_used_bytes_by_compaction_session_process(
        self, peak_memory: float
    ) -> CompactionSessionAuditInfo:
        self["peakMemoryUsedBytesCompactionSessionProcess"] = peak_memory
        return self

    def set_estimated_in_memory_size_bytes_during_discovery(
        self, memory: float
    ) -> CompactionSessionAuditInfo:
        self["estimatedInMemorySizeBytesDuringDiscovery"] = memory
        return self

    def set_hash_bucket_processed_size_bytes(
        self, size: int
    ) -> CompactionSessionAuditInfo:
        self["hashBucketProcessedSizeBytes"] = size
        return self

    def set_total_cpu_seconds(self, value: float) -> CompactionSessionAuditInfo:
        self["totalCPUSeconds"] = value
        return self

    def set_used_cpu_seconds(self, value: float) -> CompactionSessionAuditInfo:
        self["usedCPUSeconds"] = value
        return self

    def set_used_memory_gb_seconds(self, value: float) -> CompactionSessionAuditInfo:
        self["usedMemoryGBSeconds"] = value
        return self

    def set_total_memory_gb_seconds(self, value: float) -> CompactionSessionAuditInfo:
        self["totalMemoryGBSeconds"] = value
        return self

    def set_pyarrow_version(self, value: str) -> CompactionSessionAuditInfo:
        self["pyarrowVersion"] = value
        return self

    def set_compactor_version(self, value: str) -> CompactionSessionAuditInfo:
        self["compactorVersion"] = value
        return self

    # High level methods to save stats
    def save_step_stats(
        self,
        step_name: str,
        task_results: Union[
            List[HashBucketResult], List[DedupeResult], List[MaterializeResult]
        ],
        task_results_retrieved_at: float,
        invoke_time_in_seconds: float,
        task_time_in_seconds: float,
    ) -> float:
        """
        Saves the stats by calling individual setters and returns the cluster telemetry time.
        """

        self[f"{step_name}TimeInSeconds"] = task_time_in_seconds
        self[f"{step_name}InvokeTimeInSeconds"] = invoke_time_in_seconds

        self[f"{step_name}ResultSize"] = get_size_of_object_in_bytes(task_results)

        (
            cluster_utilization_after_task,
            cluster_util_after_task_latency,
        ) = timed_invocation(ClusterUtilization.get_current_cluster_utilization)

        self.set_total_cluster_object_store_memory_bytes(
            cluster_utilization_after_task.total_object_store_memory_bytes
        )
        self.set_total_cluster_memory_bytes(
            cluster_utilization_after_task.total_memory_bytes
        )
        self.set_total_object_store_memory_used_bytes(
            cluster_utilization_after_task.used_object_store_memory_bytes
        )

        self[
            f"{step_name}PostObjectStoreMemoryUsedBytes"
        ] = cluster_utilization_after_task.used_object_store_memory_bytes

        telemetry_time = 0
        if task_results:
            last_task_completed_at = max(
                result.task_completed_at for result in task_results
            )

            self[f"{step_name}ResultWaitTimeInSeconds"] = (
                task_results_retrieved_at - last_task_completed_at.item()
            )

            peak_task_memory = max(
                result.peak_memory_usage_bytes for result in task_results
            )

            telemetry_time = sum(
                result.telemetry_time_in_seconds for result in task_results
            )

            self[f"{step_name}TaskPeakMemoryUsedBytes"] = peak_task_memory.item()

        return cluster_util_after_task_latency + telemetry_time

    def save_round_completion_stats(
        self, mat_results: List[MaterializeResult], total_telemetry_time: float
    ) -> None:
        """
        This method saves all the relevant stats after all the steps are completed.
        """
        pyarrow_write_result = PyArrowWriteResult.union(
            [m.pyarrow_write_result for m in mat_results]
        )

        total_count_of_src_dfl_not_touched = sum(
            m.referenced_pyarrow_write_result.files
            if m.referenced_pyarrow_write_result
            else 0
            for m in mat_results
        )

        logger.info(
            f"Got total of {total_count_of_src_dfl_not_touched} manifest files not touched."
        )
        logger.info(
            f"Got total of {pyarrow_write_result.files} manifest files during compaction."
        )
        manifest_entry_copied_by_reference_ratio = (
            (
                round(
                    total_count_of_src_dfl_not_touched / pyarrow_write_result.files, 4
                )
                * 100
            )
            if pyarrow_write_result.files != 0
            else None
        )
        logger.info(
            f"{manifest_entry_copied_by_reference_ratio} percent of manifest files are copied by reference during materialize."
        )

        untouched_file_record_count = sum(
            m.referenced_pyarrow_write_result.records
            if m.referenced_pyarrow_write_result
            else 0
            for m in mat_results
        )
        untouched_file_size_bytes = sum(
            m.referenced_pyarrow_write_result.file_bytes
            if m.referenced_pyarrow_write_result
            else 0
            for m in mat_results
        )

        self.set_untouched_file_count(total_count_of_src_dfl_not_touched)
        self.set_untouched_file_ratio(manifest_entry_copied_by_reference_ratio)
        self.set_untouched_record_count(untouched_file_record_count)
        self.set_untouched_size_bytes(untouched_file_size_bytes)

        self.set_output_file_count(pyarrow_write_result.files)
        self.set_output_size_bytes(pyarrow_write_result.file_bytes)
        self.set_output_size_pyarrow_bytes(pyarrow_write_result.pyarrow_bytes)

        self.set_peak_memory_used_bytes_per_task(
            max(
                [
                    self.peak_memory_used_bytes_per_hash_bucket_task or 0,
                    self.peak_memory_used_bytes_per_dedupe_task or 0,
                    self.peak_memory_used_bytes_per_materialize_task or 0,
                    self.peak_memory_used_bytes_per_merge_task or 0,
                ]
            )
        )

        self.set_pyarrow_version(pa.__version__)
        self.set_telemetry_time_in_seconds(total_telemetry_time)



import botocore
from typing import Dict, Optional, List, Tuple
from deltacat.types.media import ContentEncoding, ContentType
from deltacat.types.partial_download import PartialParquetParameters
from deltacat.storage import (
    Manifest,
    ManifestEntry,
    interface as unimplemented_deltacat_storage,
)
from deltacat.compute.compactor.model.delta_annotated import DeltaAnnotated
from deltacat.compute.compactor.model.round_completion_info import RoundCompletionInfo
from deltacat.compute.compactor_v2.utils.primary_key_index import (
    hash_group_index_to_hash_bucket_indices,
)
from deltacat.compute.compactor_v2.constants import (
    TOTAL_MEMORY_BUFFER_PERCENTAGE,
    PARQUET_TO_PYARROW_INFLATION,
)


def _get_parquet_type_params_if_exist(
    entry: ManifestEntry,
) -> Optional[PartialParquetParameters]:
    if (
        entry.meta
        and entry.meta.content_type == ContentType.PARQUET
        and entry.meta.content_encoding == ContentEncoding.IDENTITY
        and entry.meta.content_type_parameters
    ):
        for type_params in entry.meta.content_type_parameters:
            if isinstance(type_params, PartialParquetParameters):
                return type_params
    return None


def _calculate_parquet_column_size(
    type_params: PartialParquetParameters, columns: List[str]
):
    column_size = 0.0
    for rg in type_params.row_groups_to_download:
        columns_found = 0
        row_group_meta = type_params.pq_metadata.row_group(rg)
        for col in range(row_group_meta.num_columns):
            column_meta = row_group_meta.column(col)
            if column_meta.path_in_schema in columns:
                columns_found += 1
                column_size += column_meta.total_uncompressed_size
        assert columns_found == len(columns), (
            "Columns not found in the parquet data as "
            f"{columns_found} != {len(columns)}"
        )
    return column_size * PARQUET_TO_PYARROW_INFLATION


def get_task_options(
    cpu: float, memory: float, ray_custom_resources: Optional[Dict] = None
) -> Dict:

    task_opts = {"num_cpus": cpu, "memory": memory}

    if ray_custom_resources:
        task_opts["resources"] = ray_custom_resources

    task_opts["max_retries"] = 3

    # List of possible botocore exceptions are available at
    # https://github.com/boto/botocore/blob/develop/botocore/exceptions.py
    task_opts["retry_exceptions"] = [
        botocore.exceptions.ConnectionError,
        botocore.exceptions.HTTPClientError,
        ConnectionError,
        TimeoutError,
    ]

    return task_opts


def estimate_manifest_entry_size_bytes(
    entry: ManifestEntry, previous_inflation: float, **kwargs
) -> float:
    if entry.meta.source_content_length:
        return entry.meta.source_content_length

    type_params = _get_parquet_type_params_if_exist(entry=entry)

    if type_params:
        return type_params.in_memory_size_bytes * PARQUET_TO_PYARROW_INFLATION

    return entry.meta.content_length * previous_inflation


def estimate_manifest_entry_num_rows(
    entry: ManifestEntry,
    average_record_size_bytes: float,
    previous_inflation: float,
    **kwargs,
) -> int:
    if entry.meta.record_count:
        return entry.meta.record_count

    type_params = _get_parquet_type_params_if_exist(entry=entry)

    if type_params:
        return type_params.num_rows

    total_size_bytes = estimate_manifest_entry_size_bytes(
        entry=entry, previous_inflation=previous_inflation, **kwargs
    )

    return int(total_size_bytes / average_record_size_bytes)


def estimate_manifest_entry_column_size_bytes(
    entry: ManifestEntry, columns: Optional[List[str]] = None
) -> Optional[float]:
    if not columns:
        return 0

    type_params = _get_parquet_type_params_if_exist(entry=entry)

    if type_params and type_params.pq_metadata:
        return _calculate_parquet_column_size(type_params=type_params, columns=columns)

    return None


def hash_bucket_resource_options_provider(
    index: int,
    item: DeltaAnnotated,
    previous_inflation: float,
    average_record_size_bytes: float,
    primary_keys: List[str] = None,
    ray_custom_resources: Optional[Dict] = None,
    **kwargs,
) -> Dict:
    size_bytes = 0.0
    num_rows = 0
    total_pk_size = 0

    if not item.manifest or not item.manifest.entries:
        return {"CPU": 0.01}

    for entry in item.manifest.entries:
        entry_size = estimate_manifest_entry_size_bytes(
            entry=entry, previous_inflation=previous_inflation
        )
        num_rows += estimate_manifest_entry_num_rows(
            entry=entry,
            previous_inflation=previous_inflation,
            average_record_size_bytes=average_record_size_bytes,
        )
        size_bytes += entry_size

        if primary_keys:
            pk_size = estimate_manifest_entry_column_size_bytes(
                entry=entry,
                columns=primary_keys,
            )

            if pk_size is None:
                total_pk_size += entry_size
            else:
                total_pk_size += pk_size

    # total size + pk size + pk hash column + hash bucket index column
    # Refer to hash_bucket step for more details.
    total_memory = size_bytes + total_pk_size + num_rows * 20 + num_rows * 4

    # Consider buffer
    total_memory = total_memory * (1 + TOTAL_MEMORY_BUFFER_PERCENTAGE / 100.0)

    return get_task_options(0.01, total_memory, ray_custom_resources)


def merge_resource_options_provider(
    index: int,
    item: Tuple[int, List],
    num_hash_groups: int,
    hash_group_size_bytes: Dict[int, int],
    hash_group_num_rows: Dict[int, int],
    round_completion_info: Optional[RoundCompletionInfo] = None,
    compacted_delta_manifest: Optional[Manifest] = None,
    ray_custom_resources: Optional[Dict] = None,
    primary_keys: Optional[List[str]] = None,
    deltacat_storage=unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Optional[Dict] = {},
    **kwargs,
) -> Dict:
    hb_group_idx = item[0]

    data_size = hash_group_size_bytes.get(hb_group_idx, 0)
    num_rows = hash_group_num_rows.get(hb_group_idx, 0)

    # upper bound for pk size of incremental
    pk_size_bytes = data_size
    incremental_index_array_size = num_rows * 4

    if (
        round_completion_info
        and compacted_delta_manifest
        and round_completion_info.hb_index_to_entry_range
    ):

        previous_inflation = (
            round_completion_info.compacted_pyarrow_write_result.pyarrow_bytes
            / round_completion_info.compacted_pyarrow_write_result.file_bytes
        )
        average_record_size = (
            round_completion_info.compacted_pyarrow_write_result.pyarrow_bytes
            / round_completion_info.compacted_pyarrow_write_result.records
        )

        iterable = hash_group_index_to_hash_bucket_indices(
            hb_group_idx, round_completion_info.hash_bucket_count, num_hash_groups
        )

        for hb_idx in iterable:
            if round_completion_info.hb_index_to_entry_range.get(str(hb_idx)) is None:
                continue

            entry_start, entry_end = round_completion_info.hb_index_to_entry_range[
                str(hb_idx)
            ]
            for entry_index in range(entry_start, entry_end):
                entry = compacted_delta_manifest.entries[entry_index]

                current_entry_size = estimate_manifest_entry_size_bytes(
                    entry=entry, previous_inflation=previous_inflation
                )
                current_entry_rows = estimate_manifest_entry_num_rows(
                    entry=entry,
                    average_record_size_bytes=average_record_size,
                    previous_inflation=previous_inflation,
                )

                data_size += current_entry_size
                num_rows += current_entry_rows

                if primary_keys:
                    pk_size = estimate_manifest_entry_column_size_bytes(
                        entry=entry,
                        columns=primary_keys,
                    )

                    if pk_size is None:
                        pk_size_bytes += current_entry_size
                    else:
                        pk_size_bytes += pk_size

    # total data downloaded + primary key hash column + primary key column
    # + dict size for merge + incremental index array size
    total_memory = (
        data_size
        + pk_size_bytes
        + num_rows * 20
        + num_rows * 20
        + incremental_index_array_size
    )

    total_memory = total_memory * (1 + TOTAL_MEMORY_BUFFER_PERCENTAGE / 100.0)

    return get_task_options(0.01, total_memory, ray_custom_resources)



from typing import List
import pyarrow as pa
from deltacat.storage import Delta, Partition
import deltacat.tests.local_deltacat_storage as ds


def create_delta_from_csv_file(
    namespace: str, file_paths: List[str], *args, **kwargs
) -> Delta:
    staged_partition = stage_partition_from_file_paths(
        namespace, file_paths, *args, **kwargs
    )

    committed_delta = commit_delta_to_staged_partition(
        staged_partition, file_paths, *args, **kwargs
    )

    return committed_delta


def stage_partition_from_file_paths(
    namespace: str, file_paths: List[str], *args, **kwargs
) -> Partition:
    ds.create_namespace(namespace, {}, **kwargs)
    table_name = "-".join(file_paths).replace("/", "_")
    ds.create_table_version(namespace, table_name, "1", **kwargs)
    stream = ds.get_stream(namespace, table_name, "1", **kwargs)
    staged_partition = ds.stage_partition(stream, [], **kwargs)
    return staged_partition


def commit_delta_to_staged_partition(
    staged_partition, file_paths: List[str], *args, **kwargs
) -> Delta:
    tables = []

    for file_path in file_paths:
        table = pa.csv.read_csv(file_path)
        tables.append(table)
    deltas = []

    for table in tables:
        delta = ds.stage_delta(table, staged_partition, **kwargs)
        deltas.append(delta)

    merged_delta = Delta.merge_deltas(deltas=deltas)
    committed_delta = ds.commit_delta(merged_delta, **kwargs)
    ds.commit_partition(staged_partition, **kwargs)
    return committed_delta



from enum import Enum
from typing import Callable, Dict, Type, Union

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as papq
from ray.data.dataset import Dataset
from ray.data.read_api import (
    from_arrow,
    from_arrow_refs,
    from_numpy,
    from_pandas,
    from_pandas_refs,
)

import deltacat.storage as dcs
from deltacat.types.media import TableType
from deltacat.utils import numpy as np_utils
from deltacat.utils import pandas as pd_utils
from deltacat.utils import pyarrow as pa_utils
from deltacat.utils.ray_utils import dataset as ds_utils

TABLE_TYPE_TO_READER_FUNC: Dict[int, Callable] = {
    TableType.PYARROW_PARQUET.value: pa_utils.s3_file_to_parquet,
    TableType.PYARROW.value: pa_utils.s3_file_to_table,
    TableType.PANDAS.value: pd_utils.s3_file_to_dataframe,
    TableType.NUMPY.value: np_utils.s3_file_to_ndarray,
}

TABLE_CLASS_TO_WRITER_FUNC: Dict[
    Type[Union[dcs.LocalTable, dcs.DistributedDataset]], Callable
] = {
    pa.Table: pa_utils.table_to_file,
    pd.DataFrame: pd_utils.dataframe_to_file,
    np.ndarray: np_utils.ndarray_to_file,
    Dataset: ds_utils.dataset_to_file,
}

TABLE_CLASS_TO_SLICER_FUNC: Dict[
    Type[Union[dcs.LocalTable, dcs.DistributedDataset]], Callable
] = {
    pa.Table: pa_utils.slice_table,
    pd.DataFrame: pd_utils.slice_dataframe,
    np.ndarray: np_utils.slice_ndarray,
    Dataset: ds_utils.slice_dataset,
}

TABLE_CLASS_TO_SIZE_FUNC: Dict[
    Type[Union[dcs.LocalTable, dcs.DistributedDataset]], Callable
] = {
    pa.Table: pa_utils.table_size,
    papq.ParquetFile: pa_utils.parquet_file_size,
    pd.DataFrame: pd_utils.dataframe_size,
    np.ndarray: np_utils.ndarray_size,
    Dataset: ds_utils.dataset_size,
}

TABLE_CLASS_TO_TABLE_TYPE: Dict[Type[dcs.LocalTable], str] = {
    pa.Table: TableType.PYARROW.value,
    papq.ParquetFile: TableType.PYARROW_PARQUET.value,
    pd.DataFrame: TableType.PANDAS.value,
    np.ndarray: TableType.NUMPY.value,
}

TABLE_TYPE_TO_DATASET_CREATE_FUNC: Dict[str, Callable] = {
    TableType.PYARROW.value: from_arrow,
    TableType.PYARROW_PARQUET.value: from_arrow,
    TableType.NUMPY.value: from_numpy,
    TableType.PANDAS.value: from_pandas,
}

TABLE_TYPE_TO_DATASET_CREATE_FUNC_REFS: Dict[str, Callable] = {
    TableType.PYARROW.value: from_arrow_refs,
    TableType.PYARROW_PARQUET.value: from_arrow_refs,
    TableType.NUMPY.value: from_numpy,
    TableType.PANDAS.value: from_pandas_refs,
}


class TableWriteMode(str, Enum):
    """
    Enum controlling how a given dataset will be written to a table.

    AUTO: CREATE if the table doesn't exist, APPEND if the table exists
    without primary keys, and MERGE if the table exists with primary keys.
    CREATE: Create the table if it doesn't exist, throw an error if it does.
    APPEND: Append to the table if it exists, throw an error if it doesn't.
    REPLACE: Replace existing table contents with the data to write.
    MERGE: Insert, update, or delete records matching a given predicate.
    Updates or inserts records based on the table's primary and sort keys by
    default.
    """

    AUTO = "auto"
    CREATE = "create"
    APPEND = "append"
    REPLACE = "replace"
    MERGE = "merge"


def get_table_length(table: Union[dcs.LocalTable, dcs.DistributedDataset]) -> int:
    return len(table) if not isinstance(table, Dataset) else table.count()


def get_table_writer(table: Union[dcs.LocalTable, dcs.DistributedDataset]) -> Callable:
    table_writer_func = TABLE_CLASS_TO_WRITER_FUNC.get(type(table))
    if table_writer_func is None:
        msg = (
            f"No writer found for table type: {type(table)}.\n"
            f"Known table types: {TABLE_CLASS_TO_WRITER_FUNC.keys}"
        )
        raise ValueError(msg)
    return table_writer_func


def get_table_slicer(table: Union[dcs.LocalTable, dcs.DistributedDataset]) -> Callable:
    table_slicer_func = TABLE_CLASS_TO_SLICER_FUNC.get(type(table))
    if table_slicer_func is None:
        msg = (
            f"No slicer found for table type: {type(table)}.\n"
            f"Known table types: {TABLE_CLASS_TO_SLICER_FUNC.keys}"
        )
        raise ValueError(msg)
    return table_slicer_func



import importlib
import logging
import time
from contextlib import nullcontext
from typing import List, Optional, Tuple
from deltacat.compute.compactor_v2.model.hash_bucket_input import HashBucketInput
import numpy as np
import pyarrow as pa
import ray
from deltacat import logs
from deltacat.compute.compactor import (
    DeltaAnnotated,
    DeltaFileEnvelope,
)
from deltacat.compute.compactor.model.delta_file_envelope import DeltaFileEnvelopeGroups
from deltacat.compute.compactor_v2.model.hash_bucket_result import HashBucketResult
from deltacat.compute.compactor_v2.utils.primary_key_index import (
    group_hash_bucket_indices,
    group_by_pk_hash_bucket,
)
from deltacat.storage import interface as unimplemented_deltacat_storage
from deltacat.types.media import StorageType
from deltacat.utils.ray_utils.runtime import (
    get_current_ray_task_id,
    get_current_ray_worker_id,
)
from deltacat.utils.common import ReadKwargsProvider
from deltacat.utils.performance import timed_invocation
from deltacat.utils.metrics import emit_timer_metrics
from deltacat.utils.resources import get_current_node_peak_memory_usage_in_bytes

if importlib.util.find_spec("memray"):
    import memray

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def _read_delta_file_envelopes(
    annotated_delta: DeltaAnnotated,
    read_kwargs_provider: Optional[ReadKwargsProvider],
    deltacat_storage=unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Optional[dict] = None,
) -> Tuple[Optional[List[DeltaFileEnvelope]], int, int]:

    tables = deltacat_storage.download_delta(
        annotated_delta,
        max_parallelism=1,
        file_reader_kwargs_provider=read_kwargs_provider,
        storage_type=StorageType.LOCAL,
        **deltacat_storage_kwargs,
    )
    annotations = annotated_delta.annotations
    assert (
        len(tables) == len(annotations),
        f"Unexpected Error: Length of downloaded delta manifest tables "
        f"({len(tables)}) doesn't match the length of delta manifest "
        f"annotations ({len(annotations)}).",
    )
    if not tables:
        return None, 0, 0

    delta_stream_position = annotations[0].annotation_stream_position
    delta_type = annotations[0].annotation_delta_type

    for annotation in annotations:
        assert annotation.annotation_stream_position == delta_stream_position, (
            f"Annotation stream position does not match - {annotation.annotation_stream_position} "
            f"!= {delta_stream_position}"
        )
        assert annotation.annotation_delta_type == delta_type, (
            f"Annotation delta type does not match - {annotation.annotation_delta_type} "
            f"!= {delta_type}"
        )

    delta_file_envelopes = []
    table = pa.concat_tables(tables)
    total_record_count = len(table)
    total_size_bytes = int(table.nbytes)

    delta_file = DeltaFileEnvelope.of(
        stream_position=delta_stream_position,
        delta_type=delta_type,
        table=table,
    )
    delta_file_envelopes.append(delta_file)
    return delta_file_envelopes, total_record_count, total_size_bytes


def _group_file_records_by_pk_hash_bucket(
    annotated_delta: DeltaAnnotated,
    num_hash_buckets: int,
    primary_keys: List[str],
    read_kwargs_provider: Optional[ReadKwargsProvider] = None,
    deltacat_storage=unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Optional[dict] = None,
) -> Tuple[Optional[DeltaFileEnvelopeGroups], int, int]:
    # read input parquet s3 objects into a list of delta file envelopes
    (
        delta_file_envelopes,
        total_record_count,
        total_size_bytes,
    ) = _read_delta_file_envelopes(
        annotated_delta,
        read_kwargs_provider,
        deltacat_storage,
        deltacat_storage_kwargs,
    )

    if delta_file_envelopes is None:
        return None, 0, 0

    logger.info(
        f"Read all delta file envelopes: {len(delta_file_envelopes)} "
        f"and total_size_bytes={total_size_bytes} and records={total_record_count}"
    )

    # group the data by primary key hash value
    hb_to_delta_file_envelopes = np.empty([num_hash_buckets], dtype="object")
    for dfe in delta_file_envelopes:
        logger.info("Grouping by pk hash bucket")
        start = time.monotonic()
        hash_bucket_to_table = group_by_pk_hash_bucket(
            dfe.table,
            num_hash_buckets,
            primary_keys,
        )
        group_end = time.monotonic()
        logger.info(f"Grouping took: {group_end - start}")
        for hb, table in enumerate(hash_bucket_to_table):
            if table:
                if hb_to_delta_file_envelopes[hb] is None:
                    hb_to_delta_file_envelopes[hb] = []
                hb_to_delta_file_envelopes[hb].append(
                    DeltaFileEnvelope.of(
                        stream_position=dfe.stream_position,
                        file_index=dfe.file_index,
                        delta_type=dfe.delta_type,
                        table=table,
                    )
                )
    return hb_to_delta_file_envelopes, total_record_count, total_size_bytes


def _timed_hash_bucket(input: HashBucketInput):
    task_id = get_current_ray_task_id()
    worker_id = get_current_ray_worker_id()
    with memray.Tracker(
        f"hash_bucket_{worker_id}_{task_id}.bin"
    ) if input.enable_profiler else nullcontext():
        (
            delta_file_envelope_groups,
            total_record_count,
            total_size_bytes,
        ) = _group_file_records_by_pk_hash_bucket(
            annotated_delta=input.annotated_delta,
            num_hash_buckets=input.num_hash_buckets,
            primary_keys=input.primary_keys,
            read_kwargs_provider=input.read_kwargs_provider,
            deltacat_storage=input.deltacat_storage,
            deltacat_storage_kwargs=input.deltacat_storage_kwargs,
        )
        hash_bucket_group_to_obj_id_tuple = group_hash_bucket_indices(
            hash_bucket_object_groups=delta_file_envelope_groups,
            num_buckets=input.num_hash_buckets,
            num_groups=input.num_hash_groups,
            object_store=input.object_store,
        )

        peak_memory_usage_bytes = get_current_node_peak_memory_usage_in_bytes()
        return HashBucketResult(
            hash_bucket_group_to_obj_id_tuple,
            np.int64(total_size_bytes),
            np.int64(total_record_count),
            np.double(peak_memory_usage_bytes),
            np.double(0.0),
            np.double(time.time()),
        )


@ray.remote
def hash_bucket(input: HashBucketInput) -> HashBucketResult:

    logger.info(f"Starting hash bucket task...")
    hash_bucket_result, duration = timed_invocation(
        func=_timed_hash_bucket, input=input
    )

    emit_metrics_time = 0.0
    if input.metrics_config:
        emit_result, latency = timed_invocation(
            func=emit_timer_metrics,
            metrics_name="hash_bucket",
            value=duration,
            metrics_config=input.metrics_config,
        )
        emit_metrics_time = latency

    logger.info(f"Finished hash bucket task...")
    return HashBucketResult(
        hash_bucket_result[0],
        hash_bucket_result[1],
        hash_bucket_result[2],
        hash_bucket_result[3],
        np.double(emit_metrics_time),
        hash_bucket_result[5],
    )



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

from collections import defaultdict
from typing import Any, Dict, List, NamedTuple, Optional, Set

from deltacat.compute.stats.models.delta_column_stats import DeltaColumnStats
from deltacat.compute.stats.models.manifest_entry_stats import ManifestEntryStats
from deltacat.compute.stats.models.stats_result import StatsResult
from deltacat.compute.stats.types import StatsType
from deltacat.storage import DeltaLocator


class DeltaStats(dict):
    """
    Stats container for all columns of a delta.

    Provides distinct stats for each delta manifest entry, aggregate stats across all manifest entries,
    and a DeltaColumnStats reference for each column.

    Each DeltaColumnStats has a column name and a ManifestEntryStats object,
    which contains column-level stats for each delta manifest entry.

    Example of visual representation:
        Manifest Entry 1
        =======
        foo bar baz
        A   B   C
        D   E   F

        Manifest Entry 2
        =======
        foo bar baz
        G   H   I
        J   K   L

        DeltaStats([
            DeltaColumnStats("foo",
                ManifestEntryStats([
                    StatsResult([A, D]),     #  Manifest Entry 1
                    StatsResult([G, J]),     #  Manifest Entry 2
                ]))
            DeltaColumnStats("bar",
                ManifestEntryStats([
                    StatsResult([B, E]),     #  Manifest Entry 1
                    StatsResult([H, K]),     #  Manifest Entry 2
                ]))
            DeltaColumnStats("baz",
                ManifestEntryStats([
                    StatsResult([C, F]),     #  Manifest Entry 1
                    StatsResult([I, L]),     #  Manifest Entry 2
                ]))
        ], Stats(AllDeltaColumnStats))
    """

    @staticmethod
    def of(column_stats: List[DeltaColumnStats]) -> DeltaStats:
        ds = DeltaStats()
        ds["column_stats"] = column_stats
        ds["stats"] = DeltaStats.get_delta_stats(column_stats)
        return ds

    @staticmethod
    def build_from_dict(delta_stats: dict) -> DeltaStats:
        delta_column_stats_list = []
        for dcs in delta_stats["column_stats"]:
            delta_column_stats_list.append(DeltaColumnStats.build_from_dict(dcs))
        return DeltaStats.of(delta_column_stats_list)

    @property
    def column_stats(self) -> List[DeltaColumnStats]:
        """
        Returns a list of stats associated to each column in this delta.
        """
        return self["column_stats"]

    @property
    def stats(self) -> Optional[StatsResult]:
        """Returns a StatsResult object that represents this delta, aggregated by the column stats of this delta."""
        val: Dict[str, Any] = self.get("stats")
        if val is not None and not isinstance(val, StatsResult):
            self["stats"] = val = StatsResult(val)
        elif val is None and self.column_stats:
            self["stats"] = val = DeltaStats.get_delta_stats(self.column_stats)

        return val

    @property
    def columns(self) -> List[str]:
        """Returns a list of column names associated to this delta.

        Returns:
            A list of column names
        """
        return DeltaStats.get_column_names(self.column_stats)

    def manifest_entry_stats(self, manifest_entry_idx: int) -> StatsResult:
        """Calculate the stats of a manifest entry by combining its columnar stats.

        Args:
            manifest_entry_idx: The manifest entry table to calculate stats for

        Returns:
            Stats for the manifest entry.
        """
        return StatsResult.merge(
            DeltaStats.get_manifest_entry_column_stats(
                self.column_stats, manifest_entry_idx
            ),
            record_row_count_once=True,
        )

    def manifest_entry_column_stats(self, manifest_entry_idx: int) -> List[StatsResult]:
        """Fetch a list of stats for each column in a manifest entry.

        Args:
            manifest_entry_idx: The manifest entry table to calculate stats for

        Returns:
            A list of columnar stats for the manifest entry
        """
        return DeltaStats.get_manifest_entry_column_stats(
            self.column_stats, manifest_entry_idx
        )

    @staticmethod
    def get_manifest_entry_column_stats(
        columns: List[DeltaColumnStats], manifest_entry_idx: int
    ) -> List[StatsResult]:
        """Helper method to provide a list of columnar stats for a specific manifest entry.

        Returns:
            A list of columnar stats for the manifest entry
        """
        dataset_columnar_stats_list: List[ManifestEntryStats] = [
            column.manifest_stats
            for column in columns
            if column.manifest_stats is not None
        ]
        try:
            return [
                stats.stats[manifest_entry_idx] for stats in dataset_columnar_stats_list
            ]
        except IndexError:
            sci: ManifestEntryStats = dataset_columnar_stats_list[0]
            raise ValueError(
                f"Table index {manifest_entry_idx} is not present in this dataset of {sci.delta_locator} "
                f"with manifest table count of {len(sci.stats)}"
            )

    @staticmethod
    def get_column_names(columns: List[DeltaColumnStats]) -> List[str]:
        """Helper method to get the names of each column from a list of delta column stats

        Args:
            columns: A list of delta column stats

        Returns:
            A list of column names
        """
        return [column_stats.column for column_stats in columns] if columns else []

    @staticmethod
    def get_delta_stats(
        columns: List[DeltaColumnStats], stat_types: Optional[Set[StatsType]] = None
    ) -> Optional[StatsResult]:
        """Calculate the sum of provided column stats and return it

        Args:
            columns: A list of delta column stats

        Returns:
            Stats for the calculated sum
        """
        assert columns and len(columns) > 0, (
            f"Expected columns `{columns}` of type `{type(columns)}` "
            f"to be a non-empty list of DeltaColumnStats"
        )

        assert all(
            [col.manifest_stats for col in columns]
        ), f"Expected stats completion info to be present in each item of {columns} "

        manifest_entry_count = len(columns[0].manifest_stats.stats)
        column_stats_map: Dict[str, List[Optional[StatsResult]]] = defaultdict(
            lambda: [None] * manifest_entry_count
        )

        for column_stats in columns:
            for file_idx, entry_stats in enumerate(column_stats.manifest_stats.stats):
                column_stats_map[column_stats.column][file_idx] = entry_stats

        return DeltaStats._merge_stats_from_columns_to_dataset(
            DeltaStats.get_column_names(columns),
            column_stats_map,
            manifest_entry_count,
            stat_types,
        )

    @staticmethod
    def _merge_stats_from_columns_to_dataset(
        column_names: List[str],
        column_stats: Dict[str, List[Optional[StatsResult]]],
        manifest_entries_size: int,
        stat_types: Optional[Set[StatsType]] = None,
    ) -> StatsResult:
        manifest_entry_stats_summary_list: List[StatsResult] = []
        for manifest_entry_idx in range(manifest_entries_size):
            curr_manifest_entry_column_stats_list: List[StatsResult] = []
            for column_name in column_names:
                current_table_column_stats: StatsResult = column_stats[column_name][
                    manifest_entry_idx
                ]
                curr_manifest_entry_column_stats_list.append(current_table_column_stats)

            curr_manifest_entry_stats_summary = StatsResult.merge(
                curr_manifest_entry_column_stats_list,
                stat_types,
                record_row_count_once=True,
            )
            manifest_entry_stats_summary_list.append(curr_manifest_entry_stats_summary)
        return StatsResult.merge(manifest_entry_stats_summary_list, stat_types)


class DeltaStatsCacheMiss(NamedTuple):
    """A helper class for cache miss results from DeltaStatsCacheResult.

    `column_names` represents missing dataset column names from the file system (ex: S3).
    delta_locator` is tied to the missing dataset columns and provided for future calculations.
    """

    column_names: List[str]
    delta_locator: DeltaLocator



from typing import Dict, Tuple, NamedTuple

import numpy as np


class DedupeResult(NamedTuple):
    mat_bucket_idx_to_obj_id: Dict[int, Tuple]
    deduped_record_count: np.int64
    peak_memory_usage_bytes: np.double
    telemetry_time_in_seconds: np.double
    task_completed_at: np.double



import importlib
import logging
from typing import Optional
import time
from collections import defaultdict
from contextlib import nullcontext
from typing import Any, Dict, List, Tuple
import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
import ray

from deltacat import logs
from deltacat.compute.compactor import (
    DeltaFileEnvelope,
    DeltaFileLocator,
)
from deltacat.storage.model.sort_key import SortKey, SortOrder
from deltacat.compute.compactor.model.dedupe_result import DedupeResult
from deltacat.compute.compactor.utils import system_columns as sc
from deltacat.utils.ray_utils.runtime import (
    get_current_ray_task_id,
    get_current_ray_worker_id,
)
from deltacat.utils.performance import timed_invocation
from deltacat.utils.metrics import emit_timer_metrics, MetricsConfig
from deltacat.io.object_store import IObjectStore
from deltacat.utils.resources import get_current_node_peak_memory_usage_in_bytes

if importlib.util.find_spec("memray"):
    import memray

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))

MaterializeBucketIndex = int
DeltaFileLocatorToRecords = Dict[DeltaFileLocator, np.ndarray]
DedupeTaskIndex, PickledObjectRef = int, str
DedupeTaskIndexWithObjectId = Tuple[DedupeTaskIndex, PickledObjectRef]


def _union_primary_key_indices(
    hash_bucket_index: int, df_envelopes_list: List[List[DeltaFileEnvelope]]
) -> pa.Table:
    logger.info(
        f"[Hash bucket index {hash_bucket_index}] Reading dedupe input for "
        f"{len(df_envelopes_list)} delta file envelope lists..."
    )
    hb_tables = []
    # sort by delta file stream position now instead of sorting every row later
    df_envelopes = [d for dfe_list in df_envelopes_list for d in dfe_list]
    df_envelopes = sorted(
        df_envelopes,
        key=lambda df: (df.stream_position, df.file_index),
        reverse=False,  # ascending
    )
    for df_envelope in df_envelopes:
        hb_tables.append(sc.project_delta_file_metadata_on_table(df_envelope))

    hb_table = pa.concat_tables(hb_tables)

    logger.info(
        f"Total records in hash bucket {hash_bucket_index} is {hb_table.num_rows}"
    )
    return hb_table


def _drop_duplicates_by_primary_key_hash(table: pa.Table) -> pa.Table:
    value_to_last_row_idx = {}

    pk_hash_np = sc.pk_hash_column_np(table)
    op_type_np = sc.delta_type_column_np(table)

    assert len(pk_hash_np) == len(op_type_np), (
        f"Primary key digest column length ({len(pk_hash_np)}) doesn't "
        f"match delta type column length ({len(op_type_np)})."
    )

    # TODO(raghumdani): move the dedupe to C++ using arrow methods or similar.
    row_idx = 0
    pk_op_val_iter = zip(pk_hash_np, op_type_np)
    for (pk_val, op_val) in pk_op_val_iter:

        # operation type is True for `UPSERT` and False for `DELETE`
        if op_val:
            # UPSERT this row
            value_to_last_row_idx[pk_val] = row_idx
        else:
            # DELETE this row
            value_to_last_row_idx.pop(pk_val, None)

        row_idx += 1

    return table.take(list(value_to_last_row_idx.values()))


def delta_file_locator_to_mat_bucket_index(
    df_locator: DeltaFileLocator, materialize_bucket_count: int
) -> int:
    digest = df_locator.digest()
    return int.from_bytes(digest, "big") % materialize_bucket_count


def _timed_dedupe(
    object_ids: List[Any],
    sort_keys: List[SortKey],
    num_materialize_buckets: int,
    dedupe_task_index: int,
    enable_profiler: bool,
    object_store: Optional[IObjectStore],
    **kwargs,
):
    task_id = get_current_ray_task_id()
    worker_id = get_current_ray_worker_id()
    with memray.Tracker(
        f"dedupe_{worker_id}_{task_id}.bin"
    ) if enable_profiler else nullcontext():
        # TODO (pdames): mitigate risk of running out of memory here in cases of severe skew of primary key updates in deltas
        logger.info(
            f"[Dedupe task {dedupe_task_index}] Getting delta file envelope "
            f"groups for {len(object_ids)} object refs..."
        )
        delta_file_envelope_groups_list: List[object] = object_store.get_many(
            object_ids
        )
        hb_index_to_delta_file_envelopes_list = defaultdict(list)
        for delta_file_envelope_groups in delta_file_envelope_groups_list:
            for hb_idx, dfes in enumerate(delta_file_envelope_groups):
                if dfes is not None:
                    hb_index_to_delta_file_envelopes_list[hb_idx].append(dfes)
        src_file_id_to_row_indices = defaultdict(list)
        deduped_tables = []
        logger.info(
            f"[Dedupe task {dedupe_task_index}] Running {len(hb_index_to_delta_file_envelopes_list)} "
            f"dedupe rounds..."
        )
        total_deduped_records = 0
        for hb_idx, dfe_list in hb_index_to_delta_file_envelopes_list.items():
            logger.info(
                f"{dedupe_task_index}: union primary keys for hb_index: {hb_idx}"
            )

            table, union_time = timed_invocation(
                func=_union_primary_key_indices,
                hash_bucket_index=hb_idx,
                df_envelopes_list=dfe_list,
            )
            logger.info(
                f"[Dedupe {dedupe_task_index}] Dedupe round input "
                f"record count: {len(table)}, took {union_time}s"
            )

            # sort by sort keys
            if len(sort_keys):
                # TODO (pdames): convert to O(N) dedupe w/ sort keys
                sort_keys.extend(
                    [
                        SortKey.of(
                            sc._PARTITION_STREAM_POSITION_COLUMN_NAME,
                            SortOrder.ASCENDING,
                        ),
                        SortKey.of(
                            sc._ORDERED_FILE_IDX_COLUMN_NAME, SortOrder.ASCENDING
                        ),
                    ]
                )
                table = table.take(pc.sort_indices(table, sort_keys=sort_keys))

            # drop duplicates by primary key hash column
            logger.info(
                f"[Dedupe task index {dedupe_task_index}] Dropping duplicates for {hb_idx}"
            )

            hb_table_record_count = len(table)
            table, drop_time = timed_invocation(
                func=_drop_duplicates_by_primary_key_hash,
                table=table,
            )
            deduped_record_count = hb_table_record_count - len(table)
            total_deduped_records += deduped_record_count

            logger.info(
                f"[Dedupe task index {dedupe_task_index}] Dedupe round output "
                f"record count: {len(table)}, took: {drop_time}s"
            )

            deduped_tables.append((hb_idx, table))

            stream_position_col = sc.stream_position_column_np(table)
            file_idx_col = sc.file_index_column_np(table)
            row_idx_col = sc.record_index_column_np(table)
            is_source_col = sc.is_source_column_np(table)
            file_record_count_col = sc.file_record_count_column_np(table)
            for row_idx in range(len(table)):
                src_dfl = DeltaFileLocator.of(
                    is_source_col[row_idx],
                    stream_position_col[row_idx],
                    file_idx_col[row_idx],
                    file_record_count_col[row_idx],
                )
                # TODO(pdames): merge contiguous record number ranges
                src_file_id_to_row_indices[src_dfl].append(row_idx_col[row_idx])

        logger.info(f"Finished all dedupe rounds...")
        mat_bucket_to_src_file_records: Dict[
            MaterializeBucketIndex, DeltaFileLocatorToRecords
        ] = defaultdict(dict)
        for src_dfl, src_row_indices in src_file_id_to_row_indices.items():
            mat_bucket = delta_file_locator_to_mat_bucket_index(
                src_dfl,
                num_materialize_buckets,
            )
            mat_bucket_to_src_file_records[mat_bucket][src_dfl] = np.array(
                src_row_indices,
            )

        mat_bucket_to_dd_idx_obj_id: Dict[
            MaterializeBucketIndex, DedupeTaskIndexWithObjectId
        ] = {}
        for mat_bucket, src_file_records in mat_bucket_to_src_file_records.items():
            object_ref = object_store.put(src_file_records)
            mat_bucket_to_dd_idx_obj_id[mat_bucket] = (
                dedupe_task_index,
                object_ref,
            )
            del object_ref
        logger.info(
            f"Count of materialize buckets with object refs: "
            f"{len(mat_bucket_to_dd_idx_obj_id)}"
        )

        peak_memory_usage_bytes = get_current_node_peak_memory_usage_in_bytes()
        return DedupeResult(
            mat_bucket_to_dd_idx_obj_id,
            np.int64(total_deduped_records),
            np.double(peak_memory_usage_bytes),
            np.double(0.0),
            np.double(time.time()),
        )


@ray.remote
def dedupe(
    object_ids: List[Any],
    sort_keys: List[SortKey],
    num_materialize_buckets: int,
    dedupe_task_index: int,
    enable_profiler: bool,
    metrics_config: MetricsConfig,
    object_store: Optional[IObjectStore],
    **kwargs,
) -> DedupeResult:
    logger.info(f"[Dedupe task {dedupe_task_index}] Starting dedupe task...")
    dedupe_result, duration = timed_invocation(
        func=_timed_dedupe,
        object_ids=object_ids,
        sort_keys=sort_keys,
        num_materialize_buckets=num_materialize_buckets,
        dedupe_task_index=dedupe_task_index,
        enable_profiler=enable_profiler,
        object_store=object_store,
        **kwargs,
    )

    emit_metrics_time = 0.0
    if metrics_config:
        emit_result, latency = timed_invocation(
            func=emit_timer_metrics,
            metrics_name="dedupe",
            value=duration,
            metrics_config=metrics_config,
        )
        emit_metrics_time = latency

    logger.info(f"[Dedupe task index {dedupe_task_index}] Finished dedupe task...")
    return DedupeResult(
        dedupe_result[0],
        dedupe_result[1],
        dedupe_result[2],
        np.double(emit_metrics_time),
        dedupe_result[4],
    )



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

from collections import defaultdict
from typing import Any, Dict, List, Optional, Set

from deltacat.compute.stats.types import ALL_STATS_TYPES, StatsType


class StatsResult(dict):
    """A generic container that holds stats for a single manifest entry file."""

    @staticmethod
    def of(
        row_count: Optional[int] = 0, pyarrow_table_bytes: Optional[int] = 0
    ) -> StatsResult:
        """Static factory for building a stats result object

        Args:
            row_count: The total number of rows of a manifest entry
            pyarrow_table_bytes: The total number of bytes when loaded into memory as a PyArrow Table

        Returns:
            A stats result object
        """
        sr = StatsResult()
        sr[StatsType.ROW_COUNT.value] = row_count
        sr[StatsType.PYARROW_TABLE_BYTES.value] = pyarrow_table_bytes
        return sr

    @property
    def row_count(self) -> int:
        """Represents the row count of a manifest entry file.

        Returns:
            The total number of rows of a manifest entry
        """
        return self[StatsType.ROW_COUNT.value]

    @property
    def pyarrow_table_bytes(self) -> int:
        """Represents the size of a manifest entry file (in bytes) as it was loaded into a PyArrow table.

        Returns:
            The total number of bytes when loaded into memory as a PyArrow Table
        """
        return self[StatsType.PYARROW_TABLE_BYTES.value]

    @staticmethod
    def from_stats_types(stats_types: Dict[StatsType, Any]) -> StatsResult:
        """A helper method to filter a dictionary by supported stats and returns a stats result object.

        Args:
            stats_types: Stats that should be included for constructing a stats result

        Returns:
            A stats result object
        """
        return StatsResult(
            {
                k: v
                for k, v in stats_types.items()
                if k in [StatsType.ROW_COUNT, StatsType.PYARROW_TABLE_BYTES]
            }
        )

    @staticmethod
    def merge(
        stats_list: List[StatsResult],
        stat_types: Optional[Set[StatsType]] = None,
        record_row_count_once: bool = False,
    ) -> StatsResult:
        """Helper method to merge any list of StatsResult objects into one.

        StatsResult objects are merged by adding up their numerical stats.
        TODO (ricmiyam): Handle non-numerical stats when they are added

        Args:
            stat_types: If provided, the calculation will only include the requested stats.
            record_row_count_once: If optionally set to `True`, then row counts are only added
                from the first stats entry. One use case for this is merging table-centric stats
                by columns, since the row count is expected to be the same across different columns.

        Returns:
            A stats result object
        """
        assert isinstance(stats_list, list) and len(stats_list) > 0, (
            f"Expected stats list: {stats_list} of type {type(stats_list)} to be a "
            f"non-empty list of StatsResult objects."
        )

        # Fallback to all stat types if not provided
        stats_to_collect: Set = stat_types or ALL_STATS_TYPES

        merged_stats: Dict[StatsType, int] = defaultdict(int)
        for stats_result in stats_list:
            for stat_type in stats_to_collect:
                if stats_result:
                    merged_stats[stat_type.value] += stats_result[stat_type.value]

        if record_row_count_once and StatsType.ROW_COUNT in stats_to_collect:
            merged_stats[StatsType.ROW_COUNT.value] = stats_list[0].row_count

        return StatsResult.from_stats_types(merged_stats)



from deltacat.aws.redshift.model.manifest import (
    Manifest,
    ManifestAuthor,
    ManifestEntry,
    ManifestEntryList,
    ManifestMeta,
)

__all__ = [
    "Manifest",
    "ManifestAuthor",
    "ManifestEntry",
    "ManifestMeta",
    "ManifestEntryList",
]



TOTAL_BYTES_IN_SHA1_HASH = 20

PK_DELIMITER = "L6kl7u5f"

MAX_RECORDS_PER_COMPACTED_FILE = 4_000_000

# The maximum amount of delta bytes allowed in a batch.
# A single task will not process more than these many bytes
# unless a single manifest entry (non-parquet) or single row
# group (parquet) is bigger than this size.
MIN_DELTA_BYTES_IN_BATCH = 5_000_000_000

# The total number of files that can be processed in a
# batch. Hence, if there are tiny files, this value can be
# limited so that enough parallelism can be attained.
MIN_FILES_IN_BATCH = float("inf")

# The average record size in a table.
AVERAGE_RECORD_SIZE_BYTES = 1000

# Maximum parallelism for the tasks at each BSP step.
# Default is the number of vCPUs in about 128
# r5.8xlarge EC2 instances.
TASK_MAX_PARALLELISM = 4096

# The percentage of memory that needs to be allocated
# as buffer. This value will ensure the job doesn't run out
# of memory by considering buffer for uncertainities.
TOTAL_MEMORY_BUFFER_PERCENTAGE = 30

# The total size of records that will be hash bucketed at once
# Since, sorting is nlogn, we ensure that is not performed
# on a very large dataset for best performance.
MAX_SIZE_OF_RECORD_BATCH_IN_GIB = 2 * 1024 * 1024 * 1024

# Whether to drop duplicates during merge.
DROP_DUPLICATES = True

# PARQUET to PYARROW inflation multiplier
# This is the observed upper bound inflation for parquet
# size in metadata to pyarrow table size.
PARQUET_TO_PYARROW_INFLATION = 4



import logging
import functools
from deltacat.constants import PYARROW_INFLATION_MULTIPLIER
from deltacat.storage import (
    PartitionLocator,
    Delta,
    interface as unimplemented_deltacat_storage,
)
from deltacat import logs
from deltacat.compute.compactor.utils import io as io_v1
from deltacat.compute.compactor import DeltaAnnotated
from typing import Dict, List, Optional, Any
from deltacat.compute.compactor_v2.constants import (
    MIN_FILES_IN_BATCH,
    MIN_DELTA_BYTES_IN_BATCH,
)
from deltacat.compute.compactor.model.compaction_session_audit_info import (
    CompactionSessionAuditInfo,
)
from deltacat.compute.compactor_v2.utils.task_options import (
    estimate_manifest_entry_size_bytes,
)
from deltacat.compute.compactor_v2.utils.content_type_params import (
    append_content_type_params,
)

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def discover_deltas(
    source_partition_locator: PartitionLocator,
    last_stream_position_to_compact: int,
    rebase_source_partition_locator: Optional[PartitionLocator] = None,
    rebase_source_partition_high_watermark: Optional[int] = None,
    rcf_high_watermark: Optional[int] = None,
    deltacat_storage=unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Optional[Dict[str, Any]] = {},
    list_deltas_kwargs: Optional[Dict[str, Any]] = {},
) -> List[Delta]:

    previous_compacted_high_watermark = (
        rebase_source_partition_high_watermark or rcf_high_watermark
    )

    delta_source_partition_locator = (
        rebase_source_partition_locator or source_partition_locator
    )

    result = []

    delta_source_incremental_deltas = io_v1._discover_deltas(
        delta_source_partition_locator,
        previous_compacted_high_watermark,
        last_stream_position_to_compact,
        deltacat_storage,
        deltacat_storage_kwargs,
        list_deltas_kwargs,
    )

    result.extend(delta_source_incremental_deltas)

    logger.info(
        f"Length of input deltas from delta source table is {len(delta_source_incremental_deltas)}"
        f" from ({previous_compacted_high_watermark}, {last_stream_position_to_compact}]"
    )

    if rebase_source_partition_locator:
        previous_compacted_deltas = io_v1._discover_deltas(
            source_partition_locator,
            None,
            None,
            deltacat_storage,
            deltacat_storage_kwargs,
            list_deltas_kwargs,
        )

        result.extend(previous_compacted_deltas)

        logger.info(
            f"Length of input deltas from previous compacted table is {len(previous_compacted_deltas)}"
            f" from ({None}, {None}]"
        )

    return result


def create_uniform_input_deltas(
    input_deltas: List[Delta],
    hash_bucket_count: int,
    compaction_audit: CompactionSessionAuditInfo,
    min_delta_bytes: Optional[float] = MIN_DELTA_BYTES_IN_BATCH,
    min_file_counts: Optional[float] = MIN_FILES_IN_BATCH,
    previous_inflation: Optional[float] = PYARROW_INFLATION_MULTIPLIER,
    enable_input_split: Optional[bool] = False,
    deltacat_storage=unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Optional[Dict[str, Any]] = {},
) -> List[DeltaAnnotated]:

    delta_bytes = 0
    delta_manifest_entries_count = 0
    estimated_da_bytes = 0
    input_da_list = []

    for delta in input_deltas:
        if enable_input_split:
            append_content_type_params(
                delta=delta,
                deltacat_storage=deltacat_storage,
                deltacat_storage_kwargs=deltacat_storage_kwargs,
            )

        manifest_entries = delta.manifest.entries
        delta_manifest_entries_count += len(manifest_entries)

        for entry_index in range(len(manifest_entries)):
            entry = manifest_entries[entry_index]
            delta_bytes += entry.meta.content_length
            estimated_da_bytes += estimate_manifest_entry_size_bytes(
                entry=entry, previous_inflation=previous_inflation
            )

        delta_annotated = DeltaAnnotated.of(delta)
        input_da_list.append(delta_annotated)

    logger.info(f"Input deltas to compact this round: " f"{len(input_da_list)}")
    logger.info(f"Input delta bytes to compact: {delta_bytes}")
    logger.info(f"Input delta files to compact: {delta_manifest_entries_count}")

    size_estimation_function = functools.partial(
        estimate_manifest_entry_size_bytes, previous_inflation=previous_inflation
    )

    rebatched_da_list = DeltaAnnotated.rebatch(
        input_da_list,
        min_delta_bytes=min_delta_bytes,
        min_file_counts=min_file_counts,
        estimation_function=size_estimation_function,
    )

    compaction_audit.set_input_size_bytes(delta_bytes)
    compaction_audit.set_input_file_count(delta_manifest_entries_count)
    compaction_audit.set_estimated_in_memory_size_bytes_during_discovery(
        estimated_da_bytes
    )

    logger.info(f"Hash bucket count: {hash_bucket_count}")
    logger.info(f"Input uniform delta count: {len(rebatched_da_list)}")

    return rebatched_da_list



from deltacat.utils.common import sha1_digest, sha1_hexdigest


class Locator:
    def canonical_string(self) -> str:
        """
        Returns a unique string for the given locator that can be used
        for equality checks (i.e. two locators are equal if they have
        the same canonical string).
        """
        raise NotImplementedError()

    def digest(self) -> bytes:
        """
        Return a digest of the given locator that can be used for
        equality checks (i.e. two locators are equal if they have the
        same digest) and uniform random hash distribution.
        """
        return sha1_digest(self.canonical_string().encode("utf-8"))

    def hexdigest(self) -> str:
        """
        Returns a hexdigest of the given locator suitable
        for use in equality (i.e. two locators are equal if they have the same
        hexdigest) and inclusion in URLs.
        """
        return sha1_hexdigest(self.canonical_string().encode("utf-8"))

    def path(self, root: str, separator: str = "/") -> str:
        """
        Returns a path for the locator of the form: "{root}/{hexdigest}", where
        the default path separator of "/" may optionally be overridden with
        any string.
        """
        return f"{root}{separator}{self.hexdigest()}"
